class ResourceTags:
    def __init__(self, tags: dict[str, str] = None):
        self.tags = {}
        if isinstance(tags, ResourceTags):
            tags = tags.to_dict()
        for key, value in tags.items():
            self[key] = value

    def __setitem__(self, key, value):
        if not isinstance(key, str) or not isinstance(value, str):
            raise ValueError("Keys and values must be strings.")
        if len(key) > 256:
            raise ValueError("Keys cannot be longer than 256 characters.")
        if len(self.tags) >= 50:
            raise ValueError("Cannot have more than 50 key-value pairs.")
        self.tags[key] = value

    def __getitem__(self, key):
        return self.tags[key]

    def __hash__(self):
        return hash(frozenset(self.tags.items()))

    def to_dict(self):
        return self.tags.copy()

    def tag_names(self):
        return self.tags.keys()

    def items(self):
        return self.tags.items()

    def keys(self):
        return self.tags.keys()

    def values(self):
        return self.tags.values()



from inflection import pluralize

from titan.data_provider import fetch_resource, list_resource
from titan.enums import ResourceType
from titan.identifiers import URN, resource_label_for_type
from titan.operations.connector import connect
from titan.resources.grant import grant_yaml


def export_resources(include: list[ResourceType] = None, exclude: list[ResourceType] = None) -> dict[str, list]:
    session = connect()
    config = {}
    for resource_type in ResourceType:
        if include and resource_type not in include:
            continue
        if exclude and resource_type in exclude:
            continue
        try:
            config.update(export_resource(session, resource_type))
        except AttributeError:
            continue
    return config


def export_resource(session, resource_type: ResourceType) -> dict[str, list]:
    resource_label = resource_label_for_type(resource_type)
    resource_names = list_resource(session, resource_label)
    if len(resource_names) == 0:
        return {}
    resources = []
    for fqn in resource_names:
        resource = fetch_resource(session, URN(resource_type, fqn, account_locator=""))
        resources.append(_format_resource_config(resource, resource_type))
    return {pluralize(resource_label): resources}


def _format_resource_config(resource: dict, resource_type: ResourceType) -> dict:
    if resource_type == ResourceType.GRANT:
        return grant_yaml(resource)
    # Sort dict based on key name
    resource = {k: resource[k] for k in sorted(resource)}
    # Put name field at the top of the dict
    first_field = {}
    if "name" in resource:
        first_field = {"name": resource.pop("name")}
    return {**first_field, **resource}



import unittest

from pyparsing import ParseException

from titan.enums import DataType
from titan.resources import Database
from titan.props import (
    AlertConditionProp,
    BoolProp,
    ArgsProp,
    EnumProp,
    EnumFlagProp,
    FlagProp,
    IdentifierProp,
    IntProp,
    StringProp,
    TagsProp,
    TimeTravelProp,
)


class TestProp(unittest.TestCase):
    def validate_identity(self, prop, sql):
        self.assertEqual(prop.render(prop.parse(sql)), sql)

    def test_prop_alert_condition(self):
        assert AlertConditionProp().parse("IF(EXISTS(SELECT 1))") == "SELECT 1"
        assert AlertConditionProp().parse("IF(EXISTS(SELECT func() from tbl))") == "SELECT func() from tbl"
        self.validate_identity(AlertConditionProp(), "IF(EXISTS( SELECT 1 ))")

    def test_prop_bool(self):
        self.assertTrue(BoolProp("foo").parse("FOO = TRUE"))
        self.assertFalse(BoolProp("bar").parse("bar = FALSE"))
        self.validate_identity(BoolProp("boolprop"), "BOOLPROP = TRUE")

    def test_prop_args(self):
        self.assertEqual(ArgsProp().parse("(id INT)"), [{"name": "id", "data_type": DataType.INT}])
        self.assertEqual(ArgsProp().parse("(somestr VARCHAR)"), [{"name": "somestr", "data_type": DataType.VARCHAR}])
        self.assertEqual(ArgsProp().parse("(floaty FLOAT8)"), [{"name": "floaty", "data_type": DataType.FLOAT8}])
        self.assertEqual(
            ArgsProp().parse("(multiple INT, columns VARCHAR)"),
            [
                {"name": "multiple", "data_type": DataType.INT},
                {"name": "columns", "data_type": DataType.VARCHAR},
            ],
        )
        self.assertEqual(
            ArgsProp().parse("(id INT, name STRING, created_at TIMESTAMP)"),
            [
                {"name": "id", "data_type": DataType.INT},
                {"name": "name", "data_type": DataType.STRING},
                {"name": "created_at", "data_type": DataType.TIMESTAMP},
            ],
        )

    def test_prop_enum(self):
        self.assertEqual(EnumProp("data_type", DataType).parse("DATA_TYPE = VARCHAR"), DataType.VARCHAR)

    def test_prop_flag(self):
        self.assertEqual(FlagProp("this is a flag").parse("this is a flag"), True)
        self.assertRaises(ParseException, lambda: FlagProp("this is another flag").parse(""))

    def test_prop_enum_flag(self):
        self.assertEqual(EnumFlagProp(DataType).parse("VARCHAR"), DataType.VARCHAR)

    def test_prop_identifier(self):
        self.assertEqual(IdentifierProp("label").parse("label = value"), "value")
        self.assertEqual(IdentifierProp("label").parse('label = "value"'), '"value"')
        self.assertEqual(IdentifierProp("label").parse('label = "schema"."table"'), '"schema"."table"')
        self.assertEqual(
            IdentifierProp("label").parse('label = database."schema"."table"'), 'database."schema"."table"'
        )
        self.assertEqual(
            IdentifierProp("request_translator").parse('request_translator = "DB"."SCHEMA".function'),
            '"DB"."SCHEMA".function',
        )

    def test_prop_int(self):
        self.assertEqual(IntProp("int_prop").parse("int_prop = 42"), 42)

    def test_prop_string(self):
        self.assertEqual(StringProp("string_prop").parse("STRING_PROP = 'quoted value'"), "quoted value")
        self.assertEqual(StringProp("multi label").parse("MULTI LABEL = VALUE"), "VALUE")

    def test_prop_tags(self):
        self.assertDictEqual(TagsProp().parse("TAG (moon_phase = 'waxing')"), {"moon_phase": "waxing"})
        self.assertDictEqual(TagsProp().parse("WITH TAG (a = 'b')"), {"a": "b"})

    def test_prop_time_travel(self):
        self.assertDictEqual(TimeTravelProp("at").parse("AT(TIMESTAMP => 123)"), {"TIMESTAMP": "123"})


class TestProps(unittest.TestCase):
    def test_props_render(self):
        db = Database(name="foo", comment="bar")
        rendered = db.props.render(db.to_dict())
        self.assertEqual(
            rendered, "DATA_RETENTION_TIME_IN_DAYS = 1 MAX_DATA_EXTENSION_TIME_IN_DAYS = 14 COMMENT = $$bar$$"
        )



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ResourceType
from ..resource_name import ResourceName
from ..scope import SchemaScope
from ..enums import AccountEdition

from ..props import (
    Props,
    QueryProp,
)


@dataclass(unsafe_hash=True)
class _AggregationPolicy(ResourceSpec):
    name: ResourceName
    body: str
    # comment: str = None
    owner: Role = "SYSADMIN"


# TODO:
# Aggregation policies have a weird construction and may require new prop types to handle comment
# CREATE [ OR REPLACE ] AGGREGATION POLICY [ IF NOT EXISTS ] <name>
#   AS () RETURNS AGGREGATION_CONSTRAINT -> <body>
#   [ COMMENT = '<string_literal>' ]


class AggregationPolicy(NamedResource, Resource):
    """
    Description:
        Represents an aggregation policy in Snowflake, which defines constraints on aggregation operations.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-aggregation-policy

    Fields:
        name (string, required): The name of the aggregation policy.
        body (string, required): The SQL expression defining the aggregation constraint.
        owner (string or Role): The owner of the aggregation policy. Defaults to "SYSADMIN".

    Python:

        ```python
        aggregation_policy = AggregationPolicy(
            name="some_aggregation_policy",
            body="AGGREGATION_CONSTRAINT(MIN_GROUP_SIZE => 5)",
            owner="SYSADMIN"
        )
        ```

    Yaml:

        ```yaml
        aggregation_policies:
          - name: some_aggregation_policy
            body: AGGREGATION_CONSTRAINT(MIN_GROUP_SIZE => 5)
            owner: SYSADMIN
        ```

    """

    edition = {AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    resource_type = ResourceType.AGGREGATION_POLICY
    props = Props(
        _start_token="AS () RETURNS AGGREGATION_CONSTRAINT",
        body=QueryProp("->"),
    )
    scope = SchemaScope()
    spec = _AggregationPolicy

    def __init__(
        self,
        name: str,
        body: str,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _AggregationPolicy = _AggregationPolicy(
            name=self._name,
            body=body,
            owner=owner,
        )



from abc import ABC

from .identifiers import FQN
from .resource_name import ResourceName


class ResourceScope(ABC):
    def fully_qualified_name(self, container, resource_name: ResourceName) -> FQN:
        raise NotImplementedError


class OrganizationScope(ResourceScope):
    def fully_qualified_name(self, _, resource_name: ResourceName) -> FQN:
        return FQN(name=resource_name)


class AccountScope(ResourceScope):
    def fully_qualified_name(self, _, resource_name: ResourceName) -> FQN:
        return FQN(name=resource_name)


class DatabaseScope(ResourceScope):
    def fully_qualified_name(self, database, resource_name: ResourceName) -> FQN:
        db = database.name if database else None
        return FQN(name=resource_name, database=db)


class SchemaScope(ResourceScope):
    def fully_qualified_name(self, schema, resource_name: ResourceName) -> FQN:
        db, sch = None, None
        if schema:
            db = schema.container.name if schema.container else None
            sch = schema.name if schema else None
        return FQN(name=resource_name, database=db, schema=sch)


class TableScope(ResourceScope):
    def fully_qualified_name(self, resource_name: ResourceName):
        raise NotImplementedError
        # return FQN(
        #     name=resource_name.upper(),
        #     database=self.database_name,
        #     schema=self.schema_name,
        #     table=self.table_name,
        # )


def resource_can_be_contained_in(resource, container):
    container_type = container.__class__.__name__
    if container_type == "ResourcePointer":
        container_type = container.resource_type.value.title()
    if (
        (isinstance(resource.scope, AccountScope) and container_type == "Account")
        or (isinstance(resource.scope, DatabaseScope) and container_type == "Database")
        or (isinstance(resource.scope, SchemaScope) and container_type == "Schema")
    ):
        return True
    return False



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .database import Database
from .role import Role
from .share import Share
from ..resource_name import ResourceName
from ..enums import AccountEdition, ParseableEnum, ResourceType
from ..props import (
    FlagProp,
    EnumListProp,
    IdentifierListProp,
    Props,
    StringProp,
)
from ..scope import AccountScope


class ObjectType(ParseableEnum):
    ACCOUNT_PARAMETERS = "ACCOUNT PARAMETERS"
    DATABASES = "DATABASES"
    INTEGRATIONS = "INTEGRATIONS"
    NETWORK_POLICIES = "NETWORK POLICIES"
    RESOURCE_MONITORS = "RESOURCE MONITORS"
    ROLES = "ROLES"
    SHARES = "SHARES"
    USERS = "USERS"
    WAREHOUSES = "WAREHOUSES"


class IntegrationType(ParseableEnum):
    SECURITY_INTEGRATIONS = "SECURITY INTEGRATIONS"
    API_INTEGRATIONS = "API INTEGRATIONS"
    STORAGE_INTEGRATIONS = "STORAGE INTEGRATIONS"
    EXTERNAL_ACCESS_INTEGRATIONS = "EXTERNAL ACCESS INTEGRATIONS"
    NOTIFICATION_INTEGRATIONS = "NOTIFICATION INTEGRATIONS"


@dataclass(unsafe_hash=True)
class _ReplicationGroup(ResourceSpec):
    name: ResourceName
    object_types: list[ObjectType]
    allowed_accounts: list[str]
    allowed_databases: list[Database] = None
    allowed_shares: list[Share] = None
    allowed_integration_types: list[IntegrationType] = None
    ignore_edition_check: bool = None
    replication_schedule: str = None
    owner: Role = "SYSADMIN"


class ReplicationGroup(NamedResource, Resource):
    """
    Description:
        A replication group in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-replication-group

    Fields:
        name (string, required): The name of the replication group.
        object_types (list, required): The object types to be replicated.
        allowed_accounts (list, required): The accounts allowed to replicate.
        allowed_databases (list): The databases allowed to replicate.
        allowed_shares (list): The shares allowed to replicate.
        allowed_integration_types (list): The integration types allowed to replicate.
        ignore_edition_check (bool): Whether to ignore the edition check.
        replication_schedule (string): The replication schedule.
        owner (string or Role): The owner of the replication group. Defaults to "SYSADMIN".

    Python:

        ```python
        replication_group = ReplicationGroup(
            name="some_replication_group",
            object_types=["DATABASES"],
            allowed_accounts=["account1", "account2"],
        )
        ```

    Yaml:

        ```yaml
        replication_groups:
          - name: some_replication_group
            object_types:
              - DATABASES
            allowed_accounts:
              - account1
              - account2
        ```
    """

    edition = {AccountEdition.ENTERPRISE, AccountEdition.BUSINESS_CRITICAL}
    resource_type = ResourceType.REPLICATION_GROUP
    props = Props(
        object_types=EnumListProp("object_types", ObjectType),
        allowed_databases=IdentifierListProp("allowed_databases"),
        allowed_shares=IdentifierListProp("allowed_shares"),
        allowed_integration_types=EnumListProp("allowed_integration_types", IntegrationType),
        allowed_accounts=IdentifierListProp("allowed_accounts"),
        ignore_edition_check=FlagProp("ignore edition check"),
        replication_schedule=StringProp("replication_schedule"),
    )
    scope = AccountScope()
    spec = _ReplicationGroup

    def __init__(
        self,
        name: str,
        object_types: list[ObjectType],
        allowed_accounts: list[str],
        allowed_databases: list[Database] = None,
        allowed_shares: list[str] = None,
        allowed_integration_types: list[IntegrationType] = None,
        ignore_edition_check: bool = None,
        replication_schedule: str = None,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ReplicationGroup = _ReplicationGroup(
            name=self._name,
            object_types=object_types,
            allowed_accounts=allowed_accounts,
            allowed_databases=allowed_databases,
            allowed_shares=allowed_shares,
            allowed_integration_types=allowed_integration_types,
            ignore_edition_check=ignore_edition_check,
            replication_schedule=replication_schedule,
            owner=owner,
        )
        self.requires(self._data.allowed_databases or [])



from dataclasses import dataclass, field
from typing import Union

from ..enums import DataType, Language, NullHandling, ResourceType, Volatility
from ..identifiers import FQN
from ..props import (
    ArgsProp,
    EnumFlagProp,
    EnumProp,
    FlagProp,
    IdentifierListProp,
    Props,
    ReturnsProp,
    StringListProp,
    StringProp,
)
from ..resource_name import ResourceName
from ..scope import SchemaScope
from .resource import Arg, NamedResource, Resource, ResourceSpec
from .role import Role


@dataclass(unsafe_hash=True)
class _JavascriptUDF(ResourceSpec):
    name: ResourceName
    args: list[Arg]
    returns: str
    as_: str
    language: Language = Language.JAVASCRIPT
    comment: str = None
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    external_access_integrations: list[str] = None
    handler: str = None
    imports: list[str] = None
    null_handling: NullHandling = None
    owner: Role = "SYSADMIN"
    packages: list[str] = None
    runtime_version: str = None
    secrets: dict[str, str] = None
    secure: bool = None
    volatility: Volatility = None


class JavascriptUDF(NamedResource, Resource):
    """
    Description:
        A Javascript user-defined function (UDF) in Snowflake allows you to define a function using the JavaScript programming language.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-function

    Fields:
        name (string, required): The name of the function.
        returns (string or DataType, required): The data type of the function's return value.
        as_ (string, required): The JavaScript code to execute when the function is called.
        args (list): The arguments that the function takes.
        comment (string): A comment for the function.
        copy_grants (bool): Specifies whether to retain the access privileges from the original function when a new function is created using CREATE OR REPLACE FUNCTION. Defaults to False.
        external_access_integrations (list): External integrations accessible by the function.
        handler (string): The entry point for the function within the JavaScript code.
        imports (list): The list of JavaScript files to import.
        null_handling (string or NullHandling): How the function handles NULL input.
        owner (string or Role): The owner of the function. Defaults to "SYSADMIN".
        packages (list): The list of npm packages that the function depends on.
        runtime_version (string): The JavaScript runtime version to use.
        secrets (dict of string to string): Key-value pairs of secrets available to the function.
        secure (bool): Specifies whether the function is secure. Defaults to False.
        volatility (string or Volatility): The volatility of the function.

    Python:

        ```python
        js_udf = JavascriptUDF(
            name="some_function",
            returns="STRING",
            as_="function(x) { return x.toUpperCase(); }",
            args=[{"name": "x", "data_type": "STRING"}],
            comment="Converts a string to uppercase",
        )
        ```

    Yaml:

        ```yaml
        functions:
          - name: some_function
            returns: STRING
            as_: function(x) { return x.toUpperCase(); }
            args:
              - name: x
                data_type: STRING
            comment: Converts a string to uppercase
        ```
    """

    resource_type = ResourceType.FUNCTION

    props = Props(
        secure=FlagProp("secure"),
        args=ArgsProp(),
        # Specifies to retain the access privileges from the original function when a new
        # function is created using CREATE OR REPLACE FUNCTION.
        # copy_grants=FlagProp("copy_grants"),
        # returns=EnumProp("returns", DataType, eq=False),
        returns=ReturnsProp("returns", eq=False),
        # not_null=BoolProp("not_null"),
        language=EnumProp("language", [Language.JAVASCRIPT], eq=False),
        null_handling=EnumFlagProp(NullHandling),
        volatility=EnumFlagProp(Volatility),
        comment=StringProp("comment"),
        as_=StringProp("as", eq=False),
    )
    scope = SchemaScope()
    spec = _JavascriptUDF

    def __init__(
        self,
        name: str,
        args: list,
        returns: str,
        as_: str,
        copy_grants: bool = False,
        owner: str = "SYSADMIN",
        secure: bool = False,
        null_handling: NullHandling = None,
        volatility: Volatility = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("language", None)
        super().__init__(name, **kwargs)
        self._data: _JavascriptUDF = _JavascriptUDF(
            name=self._name,
            returns=returns,
            as_=as_,
            copy_grants=copy_grants,
            owner=owner,
            secure=secure,
            args=args,
            null_handling=null_handling,
            volatility=volatility,
            comment=comment,
        )

    @property
    def fqn(self):
        return udf_fqn(self)


@dataclass(unsafe_hash=True)
class _PythonUDF(ResourceSpec):
    name: ResourceName
    returns: str
    runtime_version: str
    handler: str
    args: list[Arg]
    language: Language = Language.PYTHON
    as_: str = None
    comment: str = None
    copy_grants: bool = field(default_factory=None, metadata={"fetchable": False})
    external_access_integrations: list[str] = None
    imports: list[str] = None
    null_handling: NullHandling = None
    owner: Role = "SYSADMIN"
    packages: list[str] = None
    secrets: dict[str, str] = None
    secure: bool = False
    volatility: Volatility = None


class PythonUDF(NamedResource, Resource):
    """
    Description:
        A Python user-defined function (UDF) in Snowflake allows users to define their own custom functions in Python.
        These functions can be used to perform operations that are not available as standard SQL functions.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-function

    Fields:
        name (string, required): The name of the function.
        returns (string, required): The data type of the function's return value.
        runtime_version (string, required): The version of the Python runtime to use.
        handler (string, required): The name of the method to call in the Python script.
        args (list, required): A list of arguments that the function takes.
        as_ (string): The Python code to execute when the function is called.
        comment (string): A comment for the function.
        copy_grants (bool): Whether to copy grants from the existing function. Defaults to False.
        external_access_integrations (list): List of external integrations accessible by the function.
        imports (list): List of modules to import in the function.
        null_handling (NullHandling): Specifies how NULL values are handled by the function.
        owner (string or Role): The owner role of the function. Defaults to "SYSADMIN".
        packages (list): List of Python packages that the function can use.
        secrets (dict): Secrets that can be accessed by the function.
        secure (bool): Whether the function is secure. Defaults to False.
        volatility (string or Volatility): The volatility of the function.

    Python:

        ```python
        python_udf = PythonUDF(
            name="some_python_udf",
            returns="string",
            runtime_version="3.8",
            handler="process_data",
            args=[{"name": "input_data", "data_type": "string"}],
            as_="process_data_function",
            comment="This function processes data.",
            copy_grants=False,
            external_access_integrations=["s3_integration"],
            imports=["pandas", "numpy"],
            null_handling="CALLED_ON_NULL_INPUT",
            owner="SYSADMIN",
            packages=["pandas", "numpy"],
            secrets={"api_key": "secret_value"},
            secure=False,
            volatility="IMMUTABLE"
        )
        ```

    Yaml:

        ```yaml
        python_udfs:
          - name: some_python_udf
            returns: string
            runtime_version: 3.8
            handler: process_data
            args:
              - name: input_data
                data_type: string
            as_: process_data_function
            comment: This function processes data.
            copy_grants: false
            external_access_integrations:
              - s3_integration
            imports:
              - pandas
              - numpy
            null_handling: CALLED_ON_NULL_INPUT
            owner: SYSADMIN
            packages:
              - pandas
              - numpy
            secrets:
              api_key: secret_value
            secure: false
            volatility: IMMUTABLE
        ```
    """

    resource_type = ResourceType.FUNCTION
    props = Props(
        secure=FlagProp("secure"),
        args=ArgsProp(),
        # returns=EnumProp("returns", DataType, eq=False),
        returns=ReturnsProp("returns", eq=False),
        language=EnumProp("language", [Language.PYTHON], eq=False),
        null_handling=EnumFlagProp(NullHandling),
        volatility=EnumFlagProp(Volatility),
        runtime_version=StringProp("runtime_version"),
        comment=StringProp("comment"),
        imports=StringListProp("imports", parens=True),
        packages=StringListProp("packages", parens=True),
        handler=StringProp("handler"),
        external_access_integrations=IdentifierListProp("external_access_integrations", parens=True),
        as_=StringProp("as", eq=False),
    )
    scope = SchemaScope()
    spec = _PythonUDF

    def __init__(
        self,
        name: str,
        args: list,
        returns: str,
        runtime_version: str,
        handler: str,
        as_: str = None,
        comment: str = None,
        copy_grants: bool = False,
        external_access_integrations: list[str] = None,
        imports: list[str] = None,
        null_handling: NullHandling = None,
        owner: str = "SYSADMIN",
        packages: list[str] = None,
        secrets: dict[str, str] = None,
        secure: bool = None,
        volatility: Volatility = None,
        **kwargs,
    ):
        kwargs.pop("language", None)
        super().__init__(name, **kwargs)
        self._data: _PythonUDF = _PythonUDF(
            name=self._name,
            returns=returns,
            runtime_version=runtime_version,
            handler=handler,
            args=args,
            as_=as_,
            comment=comment,
            copy_grants=copy_grants,
            external_access_integrations=external_access_integrations,
            imports=imports,
            null_handling=null_handling,
            owner=owner,
            packages=packages,
            secrets=secrets,
            secure=secure,
            volatility=volatility,
        )

    @property
    def fqn(self):
        return udf_fqn(self)


def udf_fqn(udf: Union[JavascriptUDF, PythonUDF]):
    schema = udf.container
    database = schema.container if schema else None
    return FQN(
        name=udf.name,
        database=database.name if database else None,
        schema=schema.name if schema else None,
        arg_types=[str(arg["data_type"]) for arg in udf._data.args],
    )


FunctionMap = {
    Language.JAVASCRIPT: JavascriptUDF,
    Language.PYTHON: PythonUDF,
}


def _resolver(data: dict):
    return FunctionMap[Language(data["language"])]


Resource.__resolvers__[ResourceType.FUNCTION] = _resolver



from dataclasses import dataclass

from ..enums import ParseableEnum, ResourceType, WarehouseSize
from ..props import (
    BoolProp,
    EnumProp,
    IdentifierProp,
    IntProp,
    Props,
    StringProp,
    TagsProp,
)
from ..resource_name import ResourceName
from ..scope import AccountScope
from .resource import NamedResource, Resource, ResourceSpec
from .resource_monitor import ResourceMonitor
from .role import Role
from .tag import TaggableResource


class WarehouseType(ParseableEnum):
    STANDARD = "STANDARD"
    SNOWPARK_OPTIMIZED = "SNOWPARK-OPTIMIZED"


class WarehouseScalingPolicy(ParseableEnum):
    STANDARD = "STANDARD"
    ECONOMY = "ECONOMY"


@dataclass(unsafe_hash=True)
class _Warehouse(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    warehouse_type: WarehouseType = WarehouseType.STANDARD
    warehouse_size: WarehouseSize = WarehouseSize.XSMALL
    max_cluster_count: int = None
    min_cluster_count: int = None
    scaling_policy: WarehouseScalingPolicy = None
    auto_suspend: int = 600
    auto_resume: bool = True
    initially_suspended: bool = None
    resource_monitor: ResourceMonitor = None
    comment: str = None
    enable_query_acceleration: bool = False
    query_acceleration_max_scale_factor: int = None
    max_concurrency_level: int = 8
    statement_queued_timeout_in_seconds: int = 0
    statement_timeout_in_seconds: int = 172800


class Warehouse(NamedResource, TaggableResource, Resource):
    """
    Description:
        A virtual warehouse, often referred to simply as a "warehouse", is a cluster of compute resources in Snowflake. It provides the necessary CPU, memory, and temporary storage to execute SQL SELECT statements, perform DML operations such as INSERT, UPDATE, DELETE, and manage data loading and unloading.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-warehouse

    Fields:
        name (string, required): The name of the warehouse.
        owner (string): The owner of the warehouse. Defaults to "SYSADMIN".
        warehouse_type (string or WarehouseType): The type of the warehouse, either STANDARD or SNOWPARK-OPTIMIZED. Defaults to STANDARD.
        warehouse_size (string or WarehouseSize): The size of the warehouse which defines the compute and storage capacity.
        max_cluster_count (int): The maximum number of clusters for the warehouse.
        min_cluster_count (int): The minimum number of clusters for the warehouse.
        scaling_policy (string or WarehouseScalingPolicy): The policy that defines how the warehouse scales.
        auto_suspend (int): The time in seconds of inactivity after which the warehouse is automatically suspended.
        auto_resume (bool): Whether the warehouse should automatically resume when queries are submitted.
        initially_suspended (bool): Whether the warehouse should start in a suspended state.
        resource_monitor (string or ResourceMonitor): The resource monitor that tracks the warehouse's credit usage and other metrics.
        comment (string): A comment about the warehouse.
        enable_query_acceleration (bool): Whether query acceleration is enabled to improve performance.
        query_acceleration_max_scale_factor (int): The maximum scale factor for query acceleration.
        max_concurrency_level (int): The maximum number of concurrent queries that the warehouse can handle.
        statement_queued_timeout_in_seconds (int): The time in seconds a statement can be queued before it times out.
        statement_timeout_in_seconds (int): The time in seconds a statement can run before it times out.
        tags (dict): Tags for the warehouse.

    Python:

        ```python
        warehouse = Warehouse(
            name="some_warehouse",
            owner="SYSADMIN",
            warehouse_type="STANDARD",
            warehouse_size="XSMALL",
            max_cluster_count=10,
            min_cluster_count=1,
            scaling_policy="STANDARD",
            auto_suspend=600,
            auto_resume=True,
            initially_suspended=False,
            resource_monitor=None,
            comment="This is a test warehouse",
            enable_query_acceleration=False,
            query_acceleration_max_scale_factor=1,
            max_concurrency_level=8,
            statement_queued_timeout_in_seconds=0,
            statement_timeout_in_seconds=172800,
            tags={"env": "test"},
        )
        ```

    Yaml:

        ```yaml
        warehouses:
          - name: some_warehouse
            owner: SYSADMIN
            warehouse_type: STANDARD
            warehouse_size: XSMALL
            max_cluster_count: 10
            min_cluster_count: 1
            scaling_policy: STANDARD
            auto_suspend: 600
            auto_resume: true
            initially_suspended: false
            resource_monitor: null
            comment: This is a test warehouse
            enable_query_acceleration: false
            query_acceleration_max_scale_factor: 1
            max_concurrency_level: 8
            statement_queued_timeout_in_seconds: 0
            statement_timeout_in_seconds: 172800
            tags:
              env: test
        ```
    """

    resource_type = ResourceType.WAREHOUSE
    props = Props(
        _start_token="WITH",
        warehouse_type=EnumProp("warehouse_type", WarehouseType),
        warehouse_size=EnumProp("warehouse_size", WarehouseSize),
        max_cluster_count=IntProp("max_cluster_count"),
        min_cluster_count=IntProp("min_cluster_count"),
        scaling_policy=EnumProp("scaling_policy", WarehouseScalingPolicy),
        auto_suspend=IntProp("auto_suspend", alt_tokens=["NULL"]),
        auto_resume=BoolProp("auto_resume"),
        initially_suspended=BoolProp("initially_suspended"),
        resource_monitor=IdentifierProp("resource_monitor"),
        comment=StringProp("comment"),
        enable_query_acceleration=BoolProp("enable_query_acceleration"),
        query_acceleration_max_scale_factor=IntProp("query_acceleration_max_scale_factor"),
        max_concurrency_level=IntProp("max_concurrency_level"),
        statement_queued_timeout_in_seconds=IntProp("statement_queued_timeout_in_seconds"),
        statement_timeout_in_seconds=IntProp("statement_timeout_in_seconds"),
        tags=TagsProp(),
    )
    scope = AccountScope()
    spec = _Warehouse

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        warehouse_type: WarehouseType = "STANDARD",
        warehouse_size: WarehouseSize = WarehouseSize.XSMALL,
        max_cluster_count: int = None,
        min_cluster_count: int = None,
        scaling_policy: WarehouseScalingPolicy = None,
        auto_suspend: int = 600,
        auto_resume: bool = True,
        initially_suspended: bool = None,
        resource_monitor: ResourceMonitor = None,
        comment: str = None,
        enable_query_acceleration: bool = False,
        query_acceleration_max_scale_factor: int = None,
        max_concurrency_level: int = 8,
        statement_queued_timeout_in_seconds: int = 0,
        statement_timeout_in_seconds: int = 172800,
        tags: dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _Warehouse = _Warehouse(
            name=self._name,
            owner=owner,
            warehouse_type=warehouse_type,
            warehouse_size=warehouse_size,
            max_cluster_count=max_cluster_count,
            min_cluster_count=min_cluster_count,
            scaling_policy=scaling_policy,
            auto_suspend=auto_suspend,
            auto_resume=auto_resume,
            initially_suspended=initially_suspended,
            resource_monitor=resource_monitor,
            comment=comment,
            enable_query_acceleration=enable_query_acceleration,
            query_acceleration_max_scale_factor=query_acceleration_max_scale_factor,
            max_concurrency_level=max_concurrency_level,
            statement_queued_timeout_in_seconds=statement_queued_timeout_in_seconds,
            statement_timeout_in_seconds=statement_timeout_in_seconds,
        )
        self.set_tags(tags)



from typing import Optional, Union

import pyparsing as pp

from .enums import ResourceType
from .parse_primitives import FullyQualifiedIdentifier
from .resource_name import ResourceName


class FQN:
    def __init__(
        self,
        name: ResourceName,
        database: Optional[ResourceName] = None,
        schema: Optional[ResourceName] = None,
        arg_types: Optional[list] = None,
        params: Optional[dict] = None,
    ) -> None:

        if not isinstance(name, ResourceName):
            raise TypeError(f"FQN name: {name} is {type(name)}, not a ResourceName")

        if database and not isinstance(database, ResourceName):
            raise TypeError(f"FQN database: {database} is {type(database)}, not a ResourceName")

        if schema and not isinstance(schema, ResourceName):
            raise TypeError(f"FQN schema: {schema} is {type(schema)}, not a ResourceName")

        self.name = name
        self.database = database
        self.schema = schema
        self.arg_types = arg_types
        self.params = params or {}

    def __eq__(self, other):
        if not isinstance(other, FQN):
            return False
        return (
            names_are_equal(self.name, other.name)
            and names_are_equal(self.database, other.database)
            and names_are_equal(self.schema, other.schema)
            and self.arg_types == other.arg_types
            and self.params == other.params
        )

    def __hash__(self):
        return hash(
            (
                self.name,
                self.database,
                self.schema,
                tuple(self.arg_types or []),
                tuple(self.params.items()),
            )
        )

    def __str__(self):
        db = f"{self.database}." if self.database else ""
        schema = f"{self.schema}." if self.schema else ""
        arg_types = ""
        if self.arg_types is not None:
            arg_types = f"({', '.join(map(str, self.arg_types))})"
        params = "?" + _params_to_str(self.params) if self.params else ""
        return f"{db}{schema}{self.name}{arg_types}{params}"

    def __repr__(self):  # pragma: no cover

        name = getattr(self, "name", None)
        database = getattr(self, "database", None)
        schema = getattr(self, "schema", None)

        db = f", db={database}" if database else ""
        schema = f", schema={schema}" if schema else ""
        arg_types = ""
        if self.arg_types is not None:
            arg_types = f", args=({', '.join(map(str, self.arg_types))})"
        params = " ?" + _params_to_str(self.params) if self.params else ""
        return f"FQN(name={name}{db}{schema}{arg_types}{params})"


class URN:
    """
    Universal Resource Name

    An address scheme for uniquely identifying resources within a Snowflake account.

    Format
    ------

                     Resource
              Account  Type         Resource
          Org     │     │            Name     Params
        ───┴── ───┴── ──┴──        ───┴───── ───┴───────
    urn:ABC123:XYZ987:table/db.sch.sometable?param=value
                            ───┬────────────
                             Fully Qualified Name
    """

    def __init__(self, resource_type: ResourceType, fqn: FQN, account_locator: str = "") -> None:
        if not isinstance(resource_type, ResourceType):
            raise Exception(f"Invalid resource type: {resource_type}")
        self.resource_type: ResourceType = resource_type
        self.resource_label: str = resource_label_for_type(resource_type)
        self.fqn: FQN = fqn
        self.account_locator: str = account_locator
        self.organization: str = ""

    def __eq__(self, other):
        if not isinstance(other, URN):
            return False
        return (
            self.resource_type == other.resource_type
            and self.fqn == other.fqn
            and self.account_locator == other.account_locator
        )

    def __hash__(self):
        return hash((self.resource_type, self.fqn, self.account_locator))

    def __str__(self):
        return f"urn:{self.organization}:{self.account_locator}:{self.resource_label}/{self.fqn}"

    def __repr__(self):  # pragma: no cover
        org = getattr(self, "organization", "")
        acct = getattr(self, "account_locator", "")
        label = getattr(self, "resource_label", "")
        fqn = getattr(self, "fqn", "")
        return f"URN(urn:{org}:{acct}:{label}/{fqn})"

    @classmethod
    def from_resource(cls, resource, account_locator: str = ""):
        return cls(resource_type=resource.resource_type, fqn=resource.fqn, account_locator=account_locator)

    def database(self):
        if not self.fqn.database:
            raise Exception(f"URN does not have a database: {self}")
        return URN(
            resource_type=ResourceType.DATABASE,
            account_locator=self.account_locator,
            fqn=FQN(name=self.fqn.database),
        )

    def schema(self):
        if not self.fqn.schema:
            raise Exception(f"URN does not have a schema: {self}")
        return URN(
            resource_type=ResourceType.SCHEMA,
            account_locator=self.account_locator,
            fqn=FQN(name=self.fqn.schema, database=self.fqn.database),
        )


# class ResourceLocator:
#     """
#     ResourceLocator

#     A simple query language for locating resources within a Snowflake account.
#     """

#     def __init__(self, resource_key: str, locator: str) -> None:
#         self.resource_key = resource_key
#         self.locator = locator
#         self.star = self.locator == "*"

#     @classmethod
#     def from_str(cls, resource_str: str) -> "ResourceLocator":
#         """
#         Parse a resource locator string.

#         Usage
#         -----
#         Locate all resources:
#         >>> ResourceLocator.from_str("*")

#         Locate a specific resource:
#         >>> ResourceLocator.from_str("database:mydb")
#         >>> ResourceLocator.from_str("schema:mydb.my_schema")
#         >>> ResourceLocator.from_str("table:mydb.my_schema.my_table")

#         Locate all resources of a given type:
#         >>> ResourceLocator.from_str("database:*")

#         Locate all resources within a given scope:
#         >>> ResourceLocator.from_str("database:mydb.*")
#         """

#         if resource_str == "*":
#             return cls(resource_key="account", locator="*")

#         parts = resource_str.split(":")
#         if len(parts) != 2:
#             raise Exception(f"Invalid resource locator string: {resource_str}")
#         return cls(resource_key=parts[0], locator=parts[1])

#     def __str__(self):
#         return f"{self.resource_key}:{self.locator}"

#     def __repr__(self):  # pragma: no cover
#         return f"ResourceLocator(resource_key='{self.resource_key}', locator='{self.locator}')"


def parse_identifier(identifier: str, is_db_scoped=False) -> dict:
    # TODO: This needs to support periods and question marks in double quoted identifiers
    scoped_name, param_str = identifier.split("?") if "?" in identifier else (identifier, "")
    params = {}
    if param_str:
        for param in param_str.split("&"):
            k, v = param.split("=")
            params[k] = v

    arg_types = None
    if "(" in scoped_name:
        args_start = scoped_name.find("(")
        scoped_name, args_str = scoped_name[:args_start], scoped_name[args_start:]
        arg_types = [arg.strip() for arg in args_str.strip("()").split(",")]

    try:
        name_parts = list(FullyQualifiedIdentifier.parse_string(scoped_name, parse_all=True))
    except pp.ParseException:
        raise pp.ParseException(f"Failed to parse identifier: {identifier}")
    if len(name_parts) == 1:
        return {
            "name": name_parts[0],
            "params": params,
            "arg_types": arg_types,
        }
    elif len(name_parts) == 2:
        if is_db_scoped:
            return {
                "database": name_parts[0],
                "name": name_parts[1],
                "params": params,
                "arg_types": arg_types,
            }
        else:
            return {
                "schema": name_parts[0],
                "name": name_parts[1],
                "params": params,
                "arg_types": arg_types,
            }
    elif len(name_parts) == 3:
        return {
            "database": name_parts[0],
            "schema": name_parts[1],
            "name": name_parts[2],
            "params": params,
            "arg_types": arg_types,
        }
    elif len(name_parts) == 4:
        params["entity"] = name_parts[3]
        return {
            "database": name_parts[0],
            "schema": name_parts[1],
            "name": name_parts[2],
            "params": params,
            "arg_types": arg_types,
        }
    raise Exception(f"Failed to parse identifier: {identifier}")


def parse_FQN(fqn_str: str, is_db_scoped=False) -> FQN:
    identifier = parse_identifier(fqn_str, is_db_scoped=is_db_scoped)
    name = identifier.pop("name")
    database = identifier.pop("database", None)
    schema = identifier.pop("schema", None)
    return FQN(
        name=ResourceName(name),
        database=ResourceName(database) if database else None,
        schema=ResourceName(schema) if schema else None,
        **identifier,
    )


# NOTE: can't put this into identifiers.py:URN because of circular import
def parse_URN(urn_str: str) -> URN:
    parts = urn_str.split(":")
    if len(parts) != 4:
        raise Exception(f"Invalid URN string: {urn_str}")
    if parts[0] != "urn":
        raise Exception(f"Invalid URN string: {urn_str}")
    resource_label, fqn_str = parts[3].split("/", 1)
    resource_type = resource_type_for_label(resource_label)
    fqn = parse_FQN(fqn_str, is_db_scoped=(resource_label == "schema"))
    return URN(
        account_locator=parts[2],
        resource_type=resource_type,
        fqn=fqn,
    )


def _params_to_str(params: dict) -> str:
    return "&".join([f"{k.lower()}={v}" for k, v in params.items()])


def resource_label_for_type(resource_type: ResourceType) -> str:
    return str(resource_type).replace(" ", "_").lower()


def resource_type_for_label(resource_label: str) -> ResourceType:
    return ResourceType(resource_label.upper().replace("_", " "))


def names_are_equal(name1: Union[None, str, ResourceName], name2: Union[None, str, ResourceName]) -> bool:
    if name1 is None and name2 is None:
        return True
    if name1 is None or name2 is None:
        return False
    return ResourceName(name1) == ResourceName(name2)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from ..resource_name import ResourceName
from .network_rule import NetworkRule
from .role import Role
from ..enums import ResourceType
from ..scope import AccountScope

from ..props import BoolProp, Props, StringProp, IdentifierListProp


@dataclass(unsafe_hash=True)
class _ExternalAccessIntegration(ResourceSpec):
    name: ResourceName
    allowed_network_rules: list[NetworkRule]
    allowed_api_authentication_integrations: list[str] = None
    # This is a special case where this resource relies on another polymorphic resource
    allowed_authentication_secrets: list[str] = None
    enabled: bool = True
    comment: str = None
    owner: Role = "ACCOUNTADMIN"

    def __post_init__(self):
        super().__post_init__()
        if self.allowed_network_rules is not None and len(self.allowed_network_rules) < 1:
            raise ValueError("allowed_network_rules must have at least one element")

        if self.allowed_api_authentication_integrations and len(self.allowed_api_authentication_integrations) < 1:
            raise ValueError("allowed_api_authentication_integrations must have at least one element if specified")

        if self.allowed_authentication_secrets:
            if len(self.allowed_authentication_secrets) < 1:
                raise ValueError("allowed_authentication_secrets must have at least one element if specified")

            if "any" in self.allowed_authentication_secrets and len(self.allowed_authentication_secrets) > 1:
                raise ValueError("allowed_authentication_secrets must not contain 'any' if there are other secrets")

            if "none" in self.allowed_authentication_secrets and len(self.allowed_authentication_secrets) > 1:
                raise ValueError("allowed_authentication_secrets must not contain 'none' if there are other secrets")


class ExternalAccessIntegration(NamedResource, Resource):
    """
    Description:
        External Access Integrations enable code within functions and stored procedures to utilize secrets and establish connections with external networks. This resource configures the rules and secrets that can be accessed by such code.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-external-access-integration

    Fields:
        name (string, required): The name of the external access integration.
        allowed_network_rules (list, required): [NetworkRules](network_rule.md) that are allowed for this integration.
        allowed_api_authentication_integrations (list): API authentication integrations that are allowed.
        allowed_authentication_secrets (list): Authentication secrets that are allowed.
        enabled (bool): Specifies if the integration is enabled. Defaults to True.
        comment (string): An optional comment about the integration.
        owner (string or Role): The owner role of the external access integration. Defaults to "ACCOUNTADMIN".

    Python:

        ```python
        external_access_integration = ExternalAccessIntegration(
            name="some_external_access_integration",
            allowed_network_rules=["rule1", "rule2"],
            enabled=True
        )
        ```

    Yaml:

        ```yaml
        external_access_integrations:
          - name: some_external_access_integration
            allowed_network_rules:
              - rule1
              - rule2
            enabled: true
        ```
    """

    resource_type = ResourceType.EXTERNAL_ACCESS_INTEGRATION
    props = Props(
        allowed_network_rules=IdentifierListProp("allowed_network_rules", eq=True, parens=True),
        allowed_api_authentication_integrations=IdentifierListProp(
            "allowed_api_authentication_integrations", eq=True, parens=True
        ),
        allowed_authentication_secrets=IdentifierListProp("allowed_authentication_secrets", eq=True, parens=True),
        enabled=BoolProp("enabled"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _ExternalAccessIntegration

    def __init__(
        self,
        name: str,
        allowed_network_rules: list[NetworkRule] = None,
        allowed_api_authentication_integrations: list[str] = None,
        allowed_authentication_secrets: list[str] = None,
        enabled: bool = True,
        comment: str = None,
        owner: str = "ACCOUNTADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ExternalAccessIntegration = _ExternalAccessIntegration(
            name=self._name,
            allowed_network_rules=allowed_network_rules or [],
            allowed_api_authentication_integrations=allowed_api_authentication_integrations,
            allowed_authentication_secrets=allowed_authentication_secrets,
            enabled=enabled,
            comment=comment,
            owner=owner,
        )
        self.requires(self._data.allowed_network_rules)
        self.requires(self._data.allowed_authentication_secrets)



from dataclasses import dataclass, field

from .resource import Resource, ResourceSpec, NamedResource
from ..enums import ParseableEnum, ResourceType
from ..resource_name import ResourceName
from ..scope import AccountScope

from ..props import (
    BoolProp,
    StringProp,
    StringListProp,
    Props,
    IntProp,
    EnumProp,
)


class SecurityIntegrationType(ParseableEnum):
    API_AUTHENTICATION = "API_AUTHENTICATION"
    EXTERNAL_OAUTH = "EXTERNAL_OAUTH"
    OAUTH = "OAUTH"
    SAML2 = "SAML"
    SCIM = "SCIM"


class OAuthClient(ParseableEnum):
    CUSTOM = "CUSTOM"
    LOOKER = "LOOKER"
    SNOWSERVICES_INGRESS = "SNOWSERVICES_INGRESS"
    TABLEAU_DESKTOP = "TABLEAU_DESKTOP"
    TABLEAU_SERVER = "TABLEAU_SERVER"


@dataclass(unsafe_hash=True)
class _SnowflakePartnerOAuthSecurityIntegration(ResourceSpec):
    name: ResourceName
    type: SecurityIntegrationType = SecurityIntegrationType.OAUTH
    enabled: bool = True
    oauth_client: OAuthClient = None
    oauth_client_secret: str = field(default_factory=None, metadata={"fetchable": False})
    oauth_redirect_uri: str = field(default_factory=None, metadata={"fetchable": False})
    oauth_issue_refresh_tokens: bool = True
    oauth_refresh_token_validity: int = 7776000
    comment: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.oauth_client not in [OAuthClient.LOOKER, OAuthClient.TABLEAU_DESKTOP, OAuthClient.TABLEAU_SERVER]:
            raise ValueError(f"Invalid OAuth client: {self.oauth_client}")


class SnowflakePartnerOAuthSecurityIntegration(NamedResource, Resource):
    """
    Description:
        A security integration in Snowflake designed to manage external OAuth clients for authentication purposes.
        This integration supports specific OAuth clients such as Looker, Tableau Desktop, and Tableau Server.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-security-integration

    Fields:
        name (string, required): The name of the security integration.
        enabled (bool): Specifies if the security integration is enabled. Defaults to True.
        oauth_client (string or OAuthClient): The OAuth client used for authentication. Supported clients are 'LOOKER', 'TABLEAU_DESKTOP', and 'TABLEAU_SERVER'.
        oauth_client_secret (string): The secret associated with the OAuth client.
        oauth_redirect_uri (string): The redirect URI configured for the OAuth client.
        oauth_issue_refresh_tokens (bool): Indicates if refresh tokens should be issued. Defaults to True.
        oauth_refresh_token_validity (int): The validity period of the refresh token in seconds.
        comment (string): A comment about the security integration.

    Python:

        ```python
        snowflake_partner_oauth_security_integration = SnowflakePartnerOAuthSecurityIntegration(
            name="some_security_integration",
            enabled=True,
            oauth_client="LOOKER",
            oauth_client_secret="secret123",
            oauth_redirect_uri="https://example.com/oauth/callback",
            oauth_issue_refresh_tokens=True,
            oauth_refresh_token_validity=7776000,
            comment="Integration for Looker OAuth"
        )
        ```

    Yaml:

        ```yaml
        security_integrations:
          - name: some_security_integration
            enabled: true
            oauth_client: LOOKER
            oauth_client_secret: secret123
            oauth_redirect_uri: https://example.com/oauth/callback
            oauth_issue_refresh_tokens: true
            oauth_refresh_token_validity: 7776000
            comment: Integration for Looker OAuth
        ```
    """

    resource_type = ResourceType.SECURITY_INTEGRATION
    props = Props(
        type=EnumProp("type", [SecurityIntegrationType.OAUTH]),
        enabled=BoolProp("enabled"),
        oauth_client=EnumProp(
            "oauth_client", [OAuthClient.LOOKER, OAuthClient.TABLEAU_DESKTOP, OAuthClient.TABLEAU_SERVER]
        ),
        oauth_client_secret=StringProp("oauth_client_secret"),
        oauth_redirect_uri=StringProp("oauth_redirect_uri"),
        oauth_issue_refresh_tokens=BoolProp("oauth_issue_refresh_tokens"),
        oauth_refresh_token_validity=IntProp("oauth_refresh_token_validity"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _SnowflakePartnerOAuthSecurityIntegration

    def __init__(
        self,
        name: str,
        enabled: bool = True,
        oauth_client: OAuthClient = None,
        oauth_client_secret: str = None,
        oauth_redirect_uri: str = None,
        oauth_issue_refresh_tokens: bool = True,
        oauth_refresh_token_validity: int = None,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        super().__init__(name, **kwargs)
        self._data = _SnowflakePartnerOAuthSecurityIntegration(
            name=self._name,
            enabled=enabled,
            oauth_client=oauth_client,
            oauth_client_secret=oauth_client_secret,
            oauth_redirect_uri=oauth_redirect_uri,
            oauth_issue_refresh_tokens=oauth_issue_refresh_tokens,
            oauth_refresh_token_validity=oauth_refresh_token_validity,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _SnowservicesOAuthSecurityIntegration(ResourceSpec):
    name: ResourceName
    type: SecurityIntegrationType = SecurityIntegrationType.OAUTH
    oauth_client: OAuthClient = OAuthClient.SNOWSERVICES_INGRESS
    enabled: bool = True
    comment: str = None


class SnowservicesOAuthSecurityIntegration(NamedResource, Resource):
    """
    Description:
        Manages OAuth security integrations for Snowservices in Snowflake, allowing external authentication mechanisms.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-security-integration

    Fields:
        name (string, required): The name of the security integration.
        enabled (bool): Specifies if the security integration is enabled. Defaults to True.
        comment (string): A comment about the security integration.

    Python:

        ```python
        snowservices_oauth = SnowservicesOAuthSecurityIntegration(
            name="some_security_integration",
            enabled=True,
            comment="Integration for external OAuth services."
        )
        ```

    Yaml:

        ```yaml
        snowservices_oauth:
          - name: some_security_integration
            enabled: true
            comment: Integration for external OAuth services.
        ```
    """

    resource_type = ResourceType.SECURITY_INTEGRATION
    props = Props(
        type=EnumProp("type", [SecurityIntegrationType.OAUTH]),
        oauth_client=EnumProp("oauth_client", [OAuthClient.SNOWSERVICES_INGRESS]),
        enabled=BoolProp("enabled"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _SnowservicesOAuthSecurityIntegration

    def __init__(
        self,
        name: str,
        enabled: bool = True,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("oauth_client", None)
        super().__init__(name, **kwargs)
        self._data = _SnowservicesOAuthSecurityIntegration(
            name=self._name,
            enabled=enabled,
            comment=comment,
        )


@dataclass(unsafe_hash=True)
class _APIAuthenticationSecurityIntegration(ResourceSpec):
    name: ResourceName
    type: SecurityIntegrationType = SecurityIntegrationType.API_AUTHENTICATION
    auth_type: str = "OAUTH2"
    enabled: bool = True
    oauth_token_endpoint: str = None
    oauth_client_auth_method: str = "CLIENT_SECRET_POST"
    oauth_client_id: str = None
    oauth_client_secret: str = field(default_factory=None, metadata={"fetchable": False})
    oauth_grant: str = None
    oauth_access_token_validity: int = None
    oauth_allowed_scopes: list[str] = None
    comment: str = None


class APIAuthenticationSecurityIntegration(NamedResource, Resource):
    """
    Description:
        Manages API authentication security integrations in Snowflake, allowing for secure API access management.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-security-integration

    Fields:
        name (string, required): The unique name of the security integration.
        auth_type (string): The type of authentication used, typically 'OAUTH2'. Defaults to 'OAUTH2'.
        oauth_token_endpoint (string): The endpoint URL for obtaining OAuth tokens.
        oauth_client_auth_method (string): The method used for client authentication, such as 'CLIENT_SECRET_POST'.
        oauth_client_id (string): The client identifier for OAuth.
        oauth_client_secret (string): The client secret for OAuth.
        oauth_grant (string): The OAuth grant type.
        oauth_access_token_validity (int): The validity period of the OAuth access token in seconds. Defaults to 0.
        oauth_allowed_scopes (list): A list of allowed scopes for the OAuth tokens.
        enabled (bool): Indicates if the security integration is enabled. Defaults to True.
        comment (string): An optional comment about the security integration.

    Python:

        ```python
        api_auth_integration = APIAuthenticationSecurityIntegration(
            name="some_api_authentication_security_integration",
            auth_type="OAUTH2",
            oauth_token_endpoint="https://example.com/oauth/token",
            oauth_client_auth_method="CLIENT_SECRET_POST",
            oauth_client_id="your_client_id",
            oauth_client_secret="your_client_secret",
            oauth_grant="client_credentials",
            oauth_access_token_validity=3600,
            oauth_allowed_scopes=["read", "write"],
            enabled=True,
            comment="Integration for external API authentication."
        )
        ```

    Yaml:

        ```yaml
        security_integrations:
        - name: some_api_authentication_security_integration
            type: api_authentication
            auth_type: OAUTH2
            oauth_token_endpoint: https://example.com/oauth/token
            oauth_client_auth_method: CLIENT_SECRET_POST
            oauth_client_id: your_client_id
            oauth_client_secret: your_client_secret
            oauth_grant: client_credentials
            oauth_access_token_validity: 3600
            oauth_allowed_scopes: [read, write]
            enabled: true
            comment: Integration for external API authentication.
        ```
    """

    resource_type = ResourceType.SECURITY_INTEGRATION
    props = Props(
        type=EnumProp("type", [SecurityIntegrationType.API_AUTHENTICATION]),
        auth_type=StringProp("auth_type"),
        enabled=BoolProp("enabled"),
        oauth_token_endpoint=StringProp("oauth_token_endpoint"),
        oauth_client_auth_method=StringProp("oauth_client_auth_method"),
        oauth_client_id=StringProp("oauth_client_id"),
        oauth_client_secret=StringProp("oauth_client_secret"),
        oauth_grant=StringProp("oauth_grant"),
        oauth_access_token_validity=IntProp("oauth_access_token_validity"),
        oauth_allowed_scopes=StringListProp("oauth_allowed_scopes", parens=True),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _APIAuthenticationSecurityIntegration

    def __init__(
        self,
        name: str,
        auth_type: str = "OAUTH2",
        oauth_token_endpoint: str = None,
        oauth_client_auth_method: str = "CLIENT_SECRET_POST",
        oauth_client_id: str = None,
        oauth_client_secret: str = None,
        oauth_grant: str = None,
        oauth_access_token_validity: int = 0,
        oauth_allowed_scopes: list[str] = None,
        enabled: bool = True,
        comment: str = None,
        **kwargs,
    ):
        kwargs.pop("type", None)
        kwargs.pop("oauth_client", None)
        super().__init__(name, **kwargs)
        self._data: _APIAuthenticationSecurityIntegration = _APIAuthenticationSecurityIntegration(
            name=self._name,
            auth_type=auth_type,
            oauth_token_endpoint=oauth_token_endpoint,
            oauth_client_auth_method=oauth_client_auth_method,
            oauth_client_id=oauth_client_id,
            oauth_client_secret=oauth_client_secret,
            oauth_grant=oauth_grant,
            oauth_access_token_validity=oauth_access_token_validity,
            oauth_allowed_scopes=oauth_allowed_scopes or [],
            enabled=enabled,
            comment=comment,
        )


def _resolver(data: dict):
    security_integration_type = SecurityIntegrationType(data["type"])
    if security_integration_type == SecurityIntegrationType.API_AUTHENTICATION:
        return APIAuthenticationSecurityIntegration
    elif security_integration_type == SecurityIntegrationType.OAUTH:
        oauth_client = OAuthClient(data["oauth_client"])
        if oauth_client in [
            OAuthClient.LOOKER,
            OAuthClient.TABLEAU_DESKTOP,
            OAuthClient.TABLEAU_SERVER,
        ]:
            return SnowflakePartnerOAuthSecurityIntegration
        elif oauth_client == OAuthClient.CUSTOM:
            # return SnowflakeCustomOAuthSecurityIntegration
            # return None
            raise NotImplementedError
        elif oauth_client == OAuthClient.SNOWSERVICES_INGRESS:
            return SnowservicesOAuthSecurityIntegration
    return None


Resource.__resolvers__[ResourceType.SECURITY_INTEGRATION] = _resolver



from dataclasses import dataclass

from ..enums import ResourceType
from ..props import BoolProp, IntProp, Props, StringListProp, StringProp, TagsProp
from ..resource_name import ResourceName
from ..scope import AccountScope
from .resource import NamedResource, Resource, ResourceSpec
from .role import Role
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _User(ResourceSpec):
    name: ResourceName
    owner: Role = "USERADMIN"
    password: str = None
    login_name: str = None
    display_name: str = None
    first_name: str = None
    middle_name: str = None
    last_name: str = None
    email: str = None
    must_change_password: bool = False
    disabled: bool = False
    days_to_expiry: int = None
    mins_to_unlock: int = None
    default_warehouse: str = None
    default_namespace: str = None
    default_role: str = None
    default_secondary_roles: list[str] = None
    mins_to_bypass_mfa: int = None
    rsa_public_key: str = None
    rsa_public_key_2: str = None
    comment: str = None
    network_policy: str = None

    def __post_init__(self):
        super().__post_init__()
        if self.login_name is None or self.login_name == "":
            self.login_name = self.name._name.upper()
        if self.display_name is None:
            self.display_name = self.name._name


class User(NamedResource, TaggableResource, Resource):
    """
    Description:
        A user in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-user

    Fields:
        name (string, required): The name of the user.
        owner (string or Role): The owner of the user. Defaults to "USERADMIN".
        password (string): The password of the user.
        login_name (string): The login name of the user. Defaults to the name in uppercase.
        display_name (string): The display name of the user. Defaults to the name in lowercase.
        first_name (string): The first name of the user.
        middle_name (string): The middle name of the user.
        last_name (string): The last name of the user.
        email (string): The email of the user.
        must_change_password (bool): Whether the user must change their password. Defaults to False.
        disabled (bool): Whether the user is disabled. Defaults to False.
        days_to_expiry (int): The number of days until the user's password expires.
        mins_to_unlock (int): The number of minutes until the user's account is unlocked.
        default_warehouse (string): The default warehouse for the user.
        default_namespace (string): The default namespace for the user.
        default_role (string): The default role for the user.
        default_secondary_roles (list): The default secondary roles for the user.
        mins_to_bypass_mfa (int): The number of minutes until the user can bypass Multi-Factor Authentication.
        rsa_public_key (string): The RSA public key for the user.
        rsa_public_key_2 (string): The RSA public key for the user.
        comment (string): A comment for the user.
        network_policy (string): The network policy for the user.
        tags (dict): Tags for the user.

    Python:

        ```python
        user = User(
            name="some_user",
            owner="USERADMIN",
            email="some.user@example.com",
        )
        ```

    Yaml:

        ```yaml
        users:
          - name: some_user
            owner: USERADMIN
            email: some.user@example.com
        ```

    """

    resource_type = ResourceType.USER
    props = Props(
        password=StringProp("password"),
        login_name=StringProp("login_name"),
        display_name=StringProp("display_name"),
        first_name=StringProp("first_name"),
        middle_name=StringProp("middle_name"),
        last_name=StringProp("last_name"),
        email=StringProp("email"),
        must_change_password=BoolProp("must_change_password"),
        disabled=BoolProp("disabled"),
        days_to_expiry=IntProp("days_to_expiry"),
        mins_to_unlock=IntProp("mins_to_unlock"),
        default_warehouse=StringProp("default_warehouse"),
        default_namespace=StringProp("default_namespace"),
        default_role=StringProp("default_role"),
        default_secondary_roles=StringListProp("default_secondary_roles", parens=True),
        mins_to_bypass_mfa=IntProp("mins_to_bypass_mfa"),
        rsa_public_key=StringProp("rsa_public_key"),
        rsa_public_key_2=StringProp("rsa_public_key_2"),
        comment=StringProp("comment"),
        network_policy=StringProp("network_policy"),
        tags=TagsProp(),
    )
    scope = AccountScope()
    spec = _User

    def __init__(
        self,
        name: str,
        owner: str = "USERADMIN",
        password: str = None,
        login_name: str = None,
        display_name: str = None,
        first_name: str = None,
        middle_name: str = None,
        last_name: str = None,
        email: str = None,
        must_change_password: bool = False,
        disabled: bool = False,
        days_to_expiry: int = None,
        mins_to_unlock: int = None,
        default_warehouse: str = None,
        default_namespace: str = None,
        default_role: str = None,
        default_secondary_roles: list[str] = None,
        mins_to_bypass_mfa: int = None,
        rsa_public_key: str = None,
        rsa_public_key_2: str = None,
        comment: str = None,
        network_policy: str = None,
        tags: dict[str, str] = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _User = _User(
            name=self._name,
            owner=owner,
            password=password,
            login_name=login_name,
            display_name=display_name,
            first_name=first_name,
            middle_name=middle_name,
            last_name=last_name,
            email=email,
            must_change_password=must_change_password,
            disabled=disabled,
            days_to_expiry=days_to_expiry,
            mins_to_unlock=mins_to_unlock,
            default_warehouse=default_warehouse,
            default_namespace=default_namespace,
            default_role=default_role,
            default_secondary_roles=default_secondary_roles,
            mins_to_bypass_mfa=mins_to_bypass_mfa,
            rsa_public_key=rsa_public_key,
            rsa_public_key_2=rsa_public_key_2,
            comment=comment,
            network_policy=network_policy,
        )
        self.set_tags(tags)

    @property
    def owner(self):
        return self._data.owner



from enum import Enum

from .resource_name import ResourceName, attribute_is_resource_name


class Action(Enum):
    ADD = "add"
    REMOVE = "remove"
    CHANGE = "change"
    TRANSFER = "transfer"


def eq(lhs, rhs, key):
    if lhs is None or rhs is None:
        return lhs == rhs

    if attribute_is_resource_name(key):
        return ResourceName(lhs) == ResourceName(rhs)
    elif key == "args":
        # Ignore arg defaults
        def _scrub_defaults(args):
            new_args = []
            for arg in args:
                new_arg = arg.copy()
                new_arg.pop("default", None)
                new_args.append(new_arg)
            return new_args

        lhs_copy = _scrub_defaults(lhs)
        rhs_copy = _scrub_defaults(rhs)
        return lhs_copy == rhs_copy
    elif key == "columns":
        if len(lhs) != len(rhs):
            return False
        for i, lhs_col in enumerate(lhs):
            rhs_col = rhs[i]
            for col_key, lhs_value in lhs_col.items():
                if col_key not in rhs_col:
                    # raise Exception(f"Column {col_key} not found in rhs {rhs}")
                    continue
                rhs_value = rhs_col[col_key]
                if not eq(lhs_value, rhs_value, col_key):
                    return False
        return True
    else:
        return lhs == rhs


def dict_delta(original, new):
    original_keys = set(original.keys())
    new_keys = set(new.keys())

    delta = {}

    for key in original_keys - new_keys:
        delta[key] = None

    for key in original_keys & new_keys:
        if not eq(original[key], new[key], key):
            delta[key] = new[key]

    for key in new_keys - original_keys:
        delta[key] = new[key]

    return delta


def diff(original, new):
    original_keys = set(original.keys())
    new_keys = set(new.keys())

    # Resources in remote state but not in the manifest should be removed
    for key in original_keys - new_keys:
        yield Action.REMOVE, key, original[key]

    # Resources in the manifest but not in remote state should be added
    for key in new_keys - original_keys:
        if isinstance(new[key], dict) and new[key].get("_pointer", False):
            raise Exception(f"Blueprint has pointer to resource that doesn't exist or isn't visible in session: {key}")
        elif isinstance(new[key], list):
            for item in new[key]:
                yield Action.ADD, key, item
        else:
            yield Action.ADD, key, new[key]

    # Resources in both should be compared
    for key in original_keys & new_keys:
        if isinstance(original[key], dict):
            # We don't diff resource pointers
            if new[key].get("_pointer", False):
                continue

            delta = dict_delta(original[key], new[key])

            owner_attr = delta.pop("owner", None)

            for attr, value in delta.items():
                yield Action.CHANGE, key, {attr: value}

            # Force the transfer to happen after all other attribute changes
            if owner_attr:
                yield Action.TRANSFER, key, {"owner": owner_attr}

        elif isinstance(original[key], list):

            for item in original[key]:
                if item not in new[key]:
                    yield Action.REMOVE, key, item

            for item in new[key]:
                if item not in original[key]:
                    yield Action.ADD, key, item



from dataclasses import dataclass

from ..enums import ResourceType
from ..props import FlagProp, IntProp, Props, StringProp, TagsProp
from ..resource_name import ResourceName
from ..scope import AccountScope
from .resource import NamedResource, Resource, ResourceContainer, ResourcePointer, ResourceSpec
from .role import Role
from .schema import Schema
from .tag import TaggableResource


@dataclass(unsafe_hash=True)
class _Database(ResourceSpec):
    name: ResourceName
    transient: bool = False
    owner: Role = "SYSADMIN"
    data_retention_time_in_days: int = 1
    max_data_extension_time_in_days: int = 14
    default_ddl_collation: str = None
    comment: str = None


class Database(NamedResource, TaggableResource, Resource, ResourceContainer):
    """
    Description:
        Represents a database in Snowflake.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-database

    Fields:
        name (string, required): The name of the database.
        transient (bool): Specifies if the database is transient. Defaults to False.
        owner (string or Role): The owner role of the database. Defaults to "SYSADMIN".
        data_retention_time_in_days (int): The number of days to retain data. Defaults to 1.
        max_data_extension_time_in_days (int): The maximum number of days to extend data retention. Defaults to 14.
        default_ddl_collation (string): The default collation for DDL statements.
        tags (dict): A dictionary of tags associated with the database.
        comment (string): A comment describing the database.

    Python:

        ```python
        database = Database(
            name="some_database",
            transient=True,
            owner="SYSADMIN",
            data_retention_time_in_days=7,
            max_data_extension_time_in_days=28,
            default_ddl_collation="utf8",
            tags={"project": "research", "priority": "high"},
            comment="This is a database."
        )
        ```

        A database can contain schemas. In Python, you can add a schema to a database in several ways:

        By database name:
        ```python
        sch = Schema(
            name = "some_schema",
            database = "my_test_db",
        )
        ```

        By database object:
        ```python
        db = Database(name = "my_test_db")
        sch = Schema(
            name = "some_schema",
            database = db,
        )
        ```

        Or using the `add` method:
        ```python
        db = Database(name = "my_test_db")
        sch = Schema(name = "some_schema")
        db.add(sch)
        ```

    Yaml:

        ```yaml
        databases:
          - name: some_database
            transient: true
            owner: SYSADMIN
            data_retention_time_in_days: 7
            max_data_extension_time_in_days: 28
            default_ddl_collation: utf8
            tags:
              project: research
              priority: high
            comment: This is a database.
        ```

        In yaml, you can add schemas to a database using the `schemas` field:

        ```yaml
        databases:
          - name: some_database
            schemas:
              - name: another_schema
        ```

        Or by name:

        ```yaml
        databases:
          - name: some_database
        schemas:
            - name: another_schema
              database: some_database
        ```
    """

    resource_type = ResourceType.DATABASE

    props = Props(
        transient=FlagProp("transient"),
        data_retention_time_in_days=IntProp("data_retention_time_in_days"),
        max_data_extension_time_in_days=IntProp("max_data_extension_time_in_days"),
        default_ddl_collation=StringProp("default_ddl_collation"),
        tags=TagsProp(),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _Database

    def __init__(
        self,
        name: str,
        transient: bool = False,
        owner: str = "SYSADMIN",
        data_retention_time_in_days: int = 1,
        max_data_extension_time_in_days: int = 14,
        default_ddl_collation: str = None,
        tags: dict[str, str] = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data = _Database(
            name=self._name,
            transient=transient,
            owner=owner,
            data_retention_time_in_days=data_retention_time_in_days,
            max_data_extension_time_in_days=max_data_extension_time_in_days,
            default_ddl_collation=default_ddl_collation,
            comment=comment,
        )
        if self._data.name != "SNOWFLAKE":
            self.add(
                Schema(name="PUBLIC", implicit=True),
                # ResourcePointer(resource_type=ResourceType.SCHEMA, name="PUBLIC")
            )
        self.set_tags(tags)

    def schemas(self):
        return self.items(resource_type=ResourceType.SCHEMA)

    @property
    def public_schema(self) -> Schema:
        return self.find(name="PUBLIC", resource_type=ResourceType.SCHEMA)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, Arg, NamedResource
from .role import Role
from ..enums import DataType, NullHandling, Volatility, ResourceType
from ..scope import SchemaScope
from ..resource_name import ResourceName

from ..props import (
    ArgsProp,
    DictProp,
    EnumProp,
    EnumFlagProp,
    FlagProp,
    IdentifierProp,
    IntProp,
    Props,
    StringProp,
)


@dataclass(unsafe_hash=True)
class _ExternalFunction(ResourceSpec):
    name: ResourceName
    returns: DataType
    api_integration: str
    as_: str
    secure: bool = False
    args: list[Arg] = None
    not_null: bool = False
    null_handling: NullHandling = None
    volatility: Volatility = None
    comment: str = None
    headers: dict[str, str] = None
    max_batch_rows: int = None
    compression: str = None
    request_translator: str = None
    response_translator: str = None
    owner: Role = "SYSADMIN"


class ExternalFunction(NamedResource, Resource):
    """
    CREATE [ OR REPLACE ] [ SECURE ] EXTERNAL FUNCTION <name> ( [ <arg_name> <arg_data_type> ] [ , ... ] )
      RETURNS <result_data_type>
      [ [ NOT ] NULL ]
      [ { CALLED ON NULL INPUT | { RETURNS NULL ON NULL INPUT | STRICT } } ]
      [ VOLATILE | IMMUTABLE ]
      [ COMMENT = '<string_literal>' ]
      API_INTEGRATION = <api_integration_name>
      [ HEADERS = ( '<header_1>' = '<value_1>' [ , '<header_2>' = '<value_2>' ... ] ) ]
      [ MAX_BATCH_ROWS = <integer> ]
      [ COMPRESSION = <compression_type> ]
      [ REQUEST_TRANSLATOR = <request_translator_udf_name> ]
      [ RESPONSE_TRANSLATOR = <response_translator_udf_name> ]
      AS <url_of_proxy_and_resource>;
    """

    resource_type = ResourceType.EXTERNAL_FUNCTION
    props = Props(
        secure=FlagProp("secure"),
        args=ArgsProp(),
        returns=EnumProp("returns", DataType, eq=False),
        # not_null=BoolProp("not_null"),
        null_handling=EnumFlagProp(NullHandling),
        volatility=EnumFlagProp(Volatility),
        comment=StringProp("comment"),
        api_integration=StringProp("api_integration"),
        headers=DictProp("headers", parens=True),
        max_batch_rows=IntProp("max_batch_rows"),
        compression=StringProp("compression"),
        request_translator=IdentifierProp("request_translator"),
        response_translator=IdentifierProp("response_translator"),
        as_=StringProp("as", eq=False),
    )
    scope = SchemaScope()
    spec = _ExternalFunction

    def __init__(
        self,
        name: str,
        returns: DataType,
        api_integration: str,
        as_: str,
        secure: bool = False,
        args: list = None,
        not_null: bool = False,
        null_handling: NullHandling = None,
        volatility: Volatility = None,
        comment: str = None,
        headers: dict[str, str] = None,
        max_batch_rows: int = None,
        compression: str = None,
        request_translator: str = None,
        response_translator: str = None,
        owner: str = "SYSADMIN",
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _ExternalFunction = _ExternalFunction(
            name=self._name,
            returns=returns,
            api_integration=api_integration,
            as_=as_,
            secure=secure,
            args=args,
            not_null=not_null,
            null_handling=null_handling,
            volatility=volatility,
            comment=comment,
            headers=headers,
            max_batch_rows=max_batch_rows,
            compression=compression,
            request_translator=request_translator,
            response_translator=response_translator,
            owner=owner,
        )



import json
import logging
from dataclasses import dataclass
from queue import Queue
from typing import Generator, Optional, Union

import snowflake.connector

from . import data_provider, lifecycle
from .client import (
    ALREADY_EXISTS_ERR,
    DOES_NOT_EXIST_ERR,
    INVALID_GRANT_ERR,
    execute,
    reset_cache,
)
from .diff import Action, diff
from .enums import ParseableEnum, ResourceType
from .identifiers import URN, parse_identifier, resource_label_for_type
from .privs import (
    CREATE_PRIV_FOR_RESOURCE_TYPE,
    PRIVS_FOR_RESOURCE_TYPE,
    AccountPriv,
    GrantedPrivilege,
    is_ownership_priv,
    execution_role_for_priv,
)
from .resource_name import ResourceName
from .resources import Account, Database
from .resources.resource import RESOURCE_SCOPES, Resource, ResourceContainer, ResourcePointer
from .resources.tag import Tag, TaggableResource
from .scope import AccountScope, DatabaseScope, OrganizationScope, ResourceScope, SchemaScope

logger = logging.getLogger("titan")

SYNC_MODE_BLOCKLIST = [
    ResourceType.FUTURE_GRANT,
    ResourceType.GRANT,
    ResourceType.GRANT_ON_ALL,
    ResourceType.ROLE,
    ResourceType.USER,
    ResourceType.TABLE,
]


class DuplicateResourceException(Exception):
    pass


class MissingResourceException(Exception):
    pass


class MissingPrivilegeException(Exception):
    pass


class MarkedForReplacementException(Exception):
    pass


class NonConformingPlanException(Exception):
    pass


class ResourceInsertionException(Exception):
    pass


class OrphanResourceException(Exception):
    pass


class InvalidOwnerException(Exception):
    pass


class RunMode(ParseableEnum):
    CREATE_OR_UPDATE = "CREATE-OR-UPDATE"
    SYNC = "SYNC"
    SYNC_ALL = "SYNC-ALL"


@dataclass
class ResourceChange:
    action: Action
    urn: URN
    before: dict
    after: dict
    delta: dict

    def to_dict(self):
        return {
            "action": self.action.value,
            "urn": str(self.urn),
            "before": self.before,
            "after": self.after,
            "delta": self.delta,
        }


State = dict[URN, dict]
Plan = list[ResourceChange]


class Manifest:
    def __init__(self, account_locator: str = ""):
        self._account_locator = account_locator
        self._data: dict[URN, Resource] = {}
        self._refs = []

    def __getitem__(self, key: URN):
        if isinstance(key, URN):
            return self._data[key]
        else:
            raise Exception("Manifest keys must be URNs")

    def __contains__(self, key):
        if isinstance(key, URN):
            return key in self._data
        else:
            raise Exception("Manifest keys must be URNs")

    def add(self, resource: Resource):

        urn = URN.from_resource(
            account_locator=self._account_locator,
            resource=resource,
        )

        # resource_data = resource.to_dict()

        if urn in self._data:
            # if resource_data != self._data[urn]:
            logger.warning(f"Duplicate resource {urn} with conflicting data, discarding {resource}")
            return
        self._data[urn] = resource
        for ref in resource.refs:
            ref_urn = URN.from_resource(account_locator=self._account_locator, resource=ref)
            self._refs.append((urn, ref_urn))

    def get(self, key: URN, default=None):
        if isinstance(key, URN):
            return self._data.get(key, default)
        else:
            raise Exception("Manifest keys must be URNs")

    def to_dict(self):
        return {k: v.to_dict() for k, v in self._data.items()}

    @property
    def urns(self) -> list[URN]:
        return list(self._data.keys())

    @property
    def refs(self):
        return self._refs


def dump_plan(plan: Plan, format: str = "json"):
    if format == "json":
        return json.dumps([change.to_dict() for change in plan], indent=2)
    elif format == "text":
        """
        account:ABC123

        » role.transformer will be created

        + role "urn::ABC123:role/transformer" {
            + name  = "transformer"
            + owner = "SYSADMIN"
            }

        + warehouse "urn::ABC123:warehouse/transforming" {
            + name           = "transforming"
            + owner          = "SYSADMIN"
            + warehouse_type = "STANDARD"
            + warehouse_size = "LARGE"
            + auto_suspend   = 60
            }

        + grant "urn::ABC123:grant/..." {
            + priv = "USAGE"
            + on   = warehouse "transforming"
            + to   = role "transformer
            }

        + grant "urn::ABC123:grant/..." {
            + priv = "OPERATE"
            + on   = warehouse "transforming"
            + to   = role "transformer
            }
        """
        output = ""

        def _render_value(value):
            if isinstance(value, str):
                return f'"{value}"'
            return str(value)

        # green_start = "\033[36m"
        # color_end = "\033[0m"

        add_count = len([change for change in plan if change.action == Action.ADD])
        change_count = len([change for change in plan if change.action == Action.CHANGE])
        transfer_count = len([change for change in plan if change.action == Action.TRANSFER])
        remove_count = len([change for change in plan if change.action == Action.REMOVE])

        output += "\n» titan core\n"
        output += f"» Plan: {add_count} to add, {change_count} to change, {transfer_count} to transfer, {remove_count} to destroy.\n\n"

        for change in plan:
            action_marker = ""
            if change.action == Action.ADD:
                action_marker = "+"
            elif change.action in (Action.CHANGE, Action.TRANSFER):
                action_marker = "~"
            elif change.action == Action.REMOVE:
                action_marker = "-"
            output += f"{action_marker} {change.urn}"
            if change.action != Action.REMOVE:
                output += " {"
            output += "\n"
            key_lengths = [len(key) for key in change.delta.keys()]
            max_key_length = max(key_lengths) if len(key_lengths) > 0 else 0
            for key, value in change.delta.items():
                if key.startswith("_"):
                    continue
                new_value = _render_value(value)
                before_value = ""
                if key in change.before:
                    before_value = _render_value(change.before[key]) + " -> "
                output += f"  {action_marker} {key:<{max_key_length}} = {before_value}{new_value}\n"
            if change.action != Action.REMOVE:
                output += "}\n"
            output += "\n"

        return output
    else:
        raise Exception(f"Unsupported format {format}")


def print_plan(plan: Plan):
    print(dump_plan(plan, format="text"))


# def plan_sql(plan: Plan) -> list[str]:
#     sql_commands = []
#     for change in plan:
#         props = Resource.props_for_resource_type(change.urn.resource_type, change.after)
#         if change.action == Action.ADD:
#             sql_commands.append(lifecycle.create_resource(change.urn, change.after, props))
#         elif change.action == Action.CHANGE:
#             sql_commands.append(lifecycle.update_resource(change.urn, change.delta))
#         elif change.action == Action.REMOVE:
#             sql_commands.append(lifecycle.drop_resource(change.urn, change.before))
#     return sql_commands


def print_diffs(diffs):
    for action, target, deltas in diffs:
        print(f"[{action}]", target)
        for delta in deltas:
            print("\t", delta)


def _split_by_scope(
    resources: list[Resource],
) -> tuple[list[Resource], list[Resource], list[Resource], list[Resource]]:
    org_scoped = []
    acct_scoped = []
    db_scoped = []
    schema_scoped = []

    seen = set()

    def route(resource: Resource):
        """The sorting hat"""

        if id(resource) not in seen:
            if isinstance(resource.scope, OrganizationScope):
                org_scoped.append(resource)
            elif isinstance(resource.scope, AccountScope):
                acct_scoped.append(resource)
            elif isinstance(resource.scope, DatabaseScope):
                db_scoped.append(resource)
            elif isinstance(resource.scope, SchemaScope):
                schema_scoped.append(resource)
            else:
                raise Exception(f"Unsupported resource type {type(resource)}")

        seen.add(id(resource))
        if isinstance(resource, ResourceContainer):
            for item in resource.items():
                route(item)

    for resource in resources:
        root = resource
        while root.container is not None:
            root = root.container
        route(root)
    return org_scoped, acct_scoped, db_scoped, schema_scoped


def _walk(resource: Resource) -> Generator[Resource, None, None]:
    yield resource
    if isinstance(resource, ResourceContainer):
        for item in resource.items():
            yield from _walk(item)


def _raise_if_plan_would_drop_session_user(session_ctx: dict, plan: Plan):
    for change in plan:
        if change.urn.resource_type == ResourceType.USER and change.action == Action.REMOVE:
            if ResourceName(session_ctx["user"]) == ResourceName(change.urn.fqn.name):
                raise Exception("Plan would drop the current session user, which is not allowed")


def _merge_pointers(resources: list[Resource]) -> list[Resource]:
    namespace = {}
    resources = sorted(resources, key=lambda resource: isinstance(resource, ResourcePointer))

    def _merge(primary: Resource, secondary: ResourcePointer):
        if secondary.container is not None:
            if primary.container is None:
                raise Exception
            secondary.container.remove(secondary)

        for item in secondary.items():
            secondary.remove(item)
            primary.add(item)

    for resource in resources:
        if hasattr(resource, "name"):
            resource_id = (resource.resource_type, resource.name)
        else:
            resource_id = str(resource.urn)
        if isinstance(resource, ResourcePointer):
            if resource_id in namespace:
                primary = namespace[resource_id]
                _merge(primary, resource)
            else:
                namespace[resource_id] = resource
        else:
            if resource_id in namespace and namespace[resource_id] != resource:

                if resource.implicit or namespace[resource_id].implicit:
                    primary, secondary = resource, namespace[resource_id]
                    if namespace[resource_id].implicit:
                        primary, secondary = secondary, primary
                    _merge(primary, secondary)
                    continue
                raise DuplicateResourceException(f"Duplicate resource found: {resource} and {namespace[resource_id]}")
            namespace[resource_id] = resource

    return list(namespace.values())


class Blueprint:

    def __init__(
        self,
        name: Optional[str] = None,
        resources: Optional[list[Resource]] = None,
        run_mode: RunMode = RunMode.CREATE_OR_UPDATE,
        dry_run: bool = False,
        allowlist: Optional[list[ResourceType]] = None,
        **kwargs,
    ) -> None:

        account = kwargs.pop("account", None)
        database = kwargs.pop("database", None)
        schema = kwargs.pop("schema", None)
        allow_role_switching = kwargs.pop("allow_role_switching", None)
        ignore_ownership = kwargs.pop("ignore_ownership", None)
        valid_resource_types = kwargs.pop("valid_resource_types", None)

        if account is not None:
            raise Exception("account is deprecated, use add instead")
        if database is not None:
            raise Exception("database is deprecated, use add instead")
        if schema is not None:
            raise Exception("schema is deprecated, use add instead")
        if allow_role_switching is not None:
            raise Exception("Role switching must be allowed")
        if ignore_ownership is not None:
            # TODO: exception
            logger.warning("ignore_ownership is deprecated")
        if valid_resource_types is not None:
            logger.warning("valid_resource_types is deprecated")
            allowlist = valid_resource_types

        self._finalized = False
        self._staged: list[Resource] = []
        self._root: Account = None
        self._run_mode: RunMode = RunMode(run_mode)
        self._dry_run: bool = dry_run
        self._allowlist: list[ResourceType] = [ResourceType(v) for v in allowlist or []]

        if self._run_mode == RunMode.SYNC_ALL:
            logger.warning("Sync All mode is dangerous, please use with caution")
            if len(self._allowlist) == 0:
                raise Exception("Sync mode must specify an allowlist")
        elif self._run_mode == RunMode.SYNC:
            if len(self._allowlist) == 0:
                raise Exception("Sync mode must specify an allowlist")
            for resource_type in self._allowlist:
                if resource_type in SYNC_MODE_BLOCKLIST:
                    raise Exception(f"Resource type {resource_type} is not allowed in sync mode")

        self.name = name

        self.add(resources or [])

    def _raise_for_nonconforming_plan(self, plan: Plan):
        exceptions = []

        # Run Mode exceptions
        if self._run_mode == RunMode.CREATE_OR_UPDATE:
            for change in plan:
                if change.action == Action.REMOVE:
                    exceptions.append(
                        f"Create-or-update mode does not allow resources to be removed (ref: {change.urn})"
                    )
                if change.action == Action.CHANGE:
                    if "name" in change.delta:
                        exceptions.append(
                            f"Create-or-update mode does not allow renaming resources (ref: {change.urn})"
                        )

        if self._run_mode == RunMode.SYNC:
            for change in plan:
                if change.urn.resource_type in SYNC_MODE_BLOCKLIST:
                    exceptions.append(
                        f"Sync mode does not allow changes to {change.urn.resource_type} (ref: {change.urn})"
                    )

        # Valid Resource Types exceptions
        if self._allowlist:
            for change in plan:
                if change.urn.resource_type not in self._allowlist:
                    exceptions.append(f"Resource type {change.urn.resource_type} not allowed in blueprint")

        if exceptions:
            if len(exceptions) > 5:
                exception_block = "\n".join(exceptions[0:5]) + f"\n... and {len(exceptions) - 5} more"
            else:
                exception_block = "\n".join(exceptions)
            raise NonConformingPlanException("Non-conforming actions found in plan:\n" + exception_block)

    def _plan(self, remote_state: State, manifest: Manifest) -> Plan:
        config = manifest.to_dict()

        changes: Plan = []
        marked_for_replacement = set()
        for action, urn, delta in diff(remote_state, config):
            before = remote_state.get(urn, {})
            after = config.get(urn, {})

            if action == Action.CHANGE:
                if urn in marked_for_replacement:
                    continue

                resource = manifest[urn]

                # TODO: if the attr is marked as must_replace, then instead we yield a rename, add, remove
                attr = list(delta.keys())[0]
                attr_metadata = resource.spec.get_metadata(attr)

                change_requires_replacement = attr_metadata.get("triggers_replacement", False)
                change_forces_add = attr_metadata.get("forces_add", False)
                change_is_fetchable = attr_metadata.get("fetchable", True)
                change_should_be_ignored = attr in resource.lifecycle.ignore_changes

                if change_requires_replacement:
                    raise MarkedForReplacementException(f"Resource {urn} is marked for replacement due to {attr}")
                    marked_for_replacement.add(urn)
                elif change_forces_add:
                    changes.append(ResourceChange(action=Action.ADD, urn=urn, before={}, after=after, delta=delta))
                    continue
                elif not change_is_fetchable:
                    # drift on fields that aren't fetchable should be ignored
                    # TODO: throw a warning, or have a blueprint runmode that fails on this
                    continue
                elif change_should_be_ignored:
                    continue
                else:
                    changes.append(ResourceChange(action, urn, before, after, delta))
            elif action == Action.ADD:
                changes.append(ResourceChange(action=action, urn=urn, before={}, after=after, delta=delta))
            elif action == Action.REMOVE:
                changes.append(ResourceChange(action=action, urn=urn, before=before, after={}, delta={}))
            elif action == Action.TRANSFER:
                resource = manifest[urn]

                attr = list(delta.keys())[0]
                attr_metadata = resource.spec.get_metadata(attr)
                change_is_fetchable = attr_metadata.get("fetchable", True)
                change_should_be_ignored = attr in resource.lifecycle.ignore_changes
                if not change_is_fetchable:
                    continue
                if change_should_be_ignored:
                    continue
                changes.append(ResourceChange(action=action, urn=urn, before=before, after=after, delta=delta))

        for urn in marked_for_replacement:
            raise MarkedForReplacementException(f"Resource {urn} is marked for replacement")
            # changes.append(ResourceChange(action=Action.REMOVE, urn=urn, before=before, after={}, delta={}))
            # changes.append(ResourceChange(action=Action.ADD, urn=urn, before={}, after=after, delta=after))

        # Generate a list of all URNs
        resource_set = set(manifest.urns + list(remote_state.keys()))
        for ref in manifest.refs:
            resource_set.add(ref[0])
            resource_set.add(ref[1])
        # Calculate a topological sort order for the URNs
        sort_order = topological_sort(resource_set, set(manifest.refs))
        return sorted(changes, key=lambda change: sort_order[change.urn])

    def fetch_remote_state(self, session, manifest: Manifest) -> State:
        state: State = {}
        session_ctx = data_provider.fetch_session(session)

        def _normalize(urn: URN, data: dict) -> dict:
            resource_cls = Resource.resolve_resource_cls(urn.resource_type, data)
            if urn.resource_type == ResourceType.FUTURE_GRANT:
                normalized = data
            elif isinstance(data, list):
                raise Exception(f"Fetching list of {urn.resource_type} is not supported yet")
            else:
                normalized = resource_cls.defaults() | data
            return normalized

        if self._run_mode in (RunMode.SYNC, RunMode.SYNC_ALL):
            """
            In sync mode, the remote state is not just the resources that were added to the blueprint,
            but all resources that exist in Snowflake. This is limited by a few things:
            - allowlist limits the scope of what resources types are allowed in a blueprint
            - if database or schema is set, the blueprint only looks at that database or schema
            """
            if len(self._allowlist) == 0:
                raise Exception("Sync mode must specify an allowlist")

            for resource_type in self._allowlist:
                for fqn in data_provider.list_resource(session, resource_label_for_type(resource_type)):
                    urn = URN(resource_type=resource_type, fqn=fqn, account_locator=session_ctx["account_locator"])
                    data = data_provider.fetch_resource(session, urn)
                    if data is None:
                        raise Exception(f"Resource {urn} not found")
                    normalized_data = _normalize(urn, data)
                    state[urn] = normalized_data

        for urn in manifest.urns:
            data = data_provider.fetch_resource(session, urn)
            if data is not None:
                normalized_data = _normalize(urn, data)
                state[urn] = normalized_data

        # check for existence of resource refs
        for parent, reference in manifest.refs:
            if reference in manifest:
                continue

            is_public_schema = reference.resource_type == ResourceType.SCHEMA and reference.fqn.name == ResourceName(
                "PUBLIC"
            )

            try:
                data = data_provider.fetch_resource(session, reference)
            except Exception:
                data = None

            if data is None and not is_public_schema:
                logger.error(manifest.to_dict())
                raise MissingResourceException(
                    f"Resource {reference} required by {parent} not found or failed to fetch"
                )

        return state

    def _build_resource_graph(self, session_ctx: dict):
        """
        Convert the staged resources into a tree of resources
        """
        org_scoped, acct_scoped, db_scoped, schema_scoped = _split_by_scope(self._staged)
        self._staged = None

        # Create root node of the resource graph
        if len(org_scoped) > 0:
            raise Exception("Blueprint cannot contain an Account resource")
        else:
            self._root = ResourcePointer(name=session_ctx["account"], resource_type=ResourceType.ACCOUNT)

        # Merge account scoped pointers into their proper resource
        acct_scoped = _merge_pointers(acct_scoped)

        # Add all databases and other account scoped resources to the root
        for resource in acct_scoped:
            self._root.add(resource)

        # List all databases connected to root
        databases: list[Union[Database, ResourcePointer]] = self._root.items(resource_type=ResourceType.DATABASE)

        # If the user didn't stage a database, create one from session context
        if len(databases) == 0 and (len(db_scoped) + len(schema_scoped) > 0):
            if session_ctx.get("database") is None:
                raise OrphanResourceException(
                    "Blueprint is missing a database but includes resources that require a database or schema"
                )

            self._root.add(ResourcePointer(name=session_ctx["database"], resource_type=ResourceType.DATABASE))
            databases = self._root.items(resource_type=ResourceType.DATABASE)

        # Attach parentless schemas to the default database, if there is one
        for resource in db_scoped:
            if resource.container is None:
                if len(databases) == 1:
                    databases[0].add(resource)
                else:
                    raise OrphanResourceException(f"Resource {resource} has no database")

        available_scopes = {}
        for database in databases:
            database_resources = list(database.items())
            _merge_pointers(database_resources)
            for schema in database.items(resource_type=ResourceType.SCHEMA):
                available_scopes[f"{database.name}.{schema.name}"] = schema

        for resource in schema_scoped:

            if resource.container is None:
                if len(databases) == 1:
                    databases[0].public_schema.add(resource)
                else:
                    raise Exception(f"No schema for resource {repr(resource)} found")
            elif isinstance(resource.container, ResourcePointer):
                schema_pointer = resource.container

                # We have a schema-scoped resource (eg a view) that has a resource pointer for the schema. The job is to connect
                # that resource into the tree
                #
                # If the schema pointer has no database, assume it lives in the only database we have
                if schema_pointer.container is None:
                    if len(databases) == 1:
                        databases[0].add(schema_pointer)
                    else:
                        raise Exception(f"No database for resource {resource} schema={resource.container}")
                elif isinstance(schema_pointer.container, ResourcePointer):
                    expected_scope = f"{schema_pointer.container.name}.{schema_pointer.name}"
                    if expected_scope in available_scopes:
                        schema = available_scopes[expected_scope]
                        schema.add(resource)
                    else:
                        self._root.add(schema_pointer.container)

            for ref in resource.refs:
                resource_and_ref_share_scope = isinstance(ref.scope, resource.scope.__class__)
                if ref.container is None and resource_and_ref_share_scope:
                    # If a resource requires another, and that secondary resource couldn't be resolved into
                    # an existing scope, then assume it lives in the same container as the original resource
                    resource.container.add(ref)

    def _create_tag_references(self):
        """
        Tag name resolution in Snowflake is special. Tags can be referenced
        by name only. If that tag name is unique in the account, the tag will be applied.
        If the tag name is not unique, the error "does not exist or not authorized" will be raised.

        To emulate this behavior, Blueprint will attempt to look up any referenced tags by name
        """
        taggables: list[TaggableResource] = []
        tags: list[Tag] = []
        for resource in _walk(self._root):
            if isinstance(resource, TaggableResource):
                taggables.append(resource)
            elif isinstance(resource, Tag):
                tags.append(resource)

        for resource in taggables:
            new_tags = {}
            if resource._tags is None:
                continue
            for tag_name, tag_value in resource._tags.items():
                identifier = parse_identifier(tag_name)
                if "database" in identifier or "schema" in identifier:
                    new_tags[tag_name] = tag_value
                else:
                    for tag in tags:
                        if tag.name == tag_name:
                            new_tags[str(tag.fqn)] = tag_value
                            break
                    else:
                        # We couldn't resolve the tag, so just use the tag name as is
                        new_tags[tag_name] = tag_value
            resource._tags = new_tags
            tag_ref = resource.create_tag_reference()
            if tag_ref:
                self._root.add(tag_ref)

    def _create_ownership_refs(self, session_ctx):
        for resource in _walk(self._root):
            if isinstance(resource, ResourcePointer):
                continue
            if hasattr(resource._data, "owner"):

                # Misconfigured resource, owner should always be a Role
                if isinstance(resource._data.owner, str):
                    raise RuntimeError(f"Owner of {resource} is a string, {resource._data.owner}")

                # Skip Snowflake-owned system resources (like INFORMATION_SCHEMA) that are owned by blank
                if resource._data.owner.name == "":
                    continue

                # Create the ownership ref
                resource.requires(resource._data.owner)

                # If the resource owner is a custom role, either it must already exist and be in available_roles
                # or we must plan to grant it to the current session
                if resource._data.owner.name not in session_ctx["available_roles"]:
                    for role_grant in self._root.items(ResourceType.ROLE_GRANT):
                        if role_grant.role.name != resource._data.owner.name:
                            continue
                        if role_grant._data.to_role is None:
                            continue
                        if role_grant.to.name not in session_ctx["available_roles"]:
                            continue
                        resource.requires(role_grant)
                        break
                    else:
                        raise InvalidOwnerException(
                            f"Blueprint resource {resource} owner {resource._data.owner} must be granted to the current session"
                        )

    def _create_grandparent_refs(self):
        for resource in _walk(self._root):
            if isinstance(resource.scope, SchemaScope):
                resource.requires(resource.container.container)

    def _finalize(self, session_ctx):
        if self._finalized:
            raise RuntimeError("Blueprint already finalized")
        self._finalized = True
        self._build_resource_graph(session_ctx)
        self._create_tag_references()
        self._create_ownership_refs(session_ctx)
        self._create_grandparent_refs()
        for resource in _walk(self._root):
            resource._finalized = True

    def generate_manifest(self, session_ctx: dict = {}) -> Manifest:
        manifest = Manifest(account_locator=session_ctx["account_locator"])
        self._finalize(session_ctx)
        for resource in _walk(self._root):
            if isinstance(resource, Resource):
                if resource.implicit:
                    continue
                manifest.add(resource)
            else:
                raise RuntimeError(f"Unexpected object found in blueprint: {resource}")

        return manifest

    def plan(self, session) -> Plan:
        reset_cache()
        session_ctx = data_provider.fetch_session(session)
        manifest = self.generate_manifest(session_ctx)
        remote_state = self.fetch_remote_state(session, manifest)
        try:
            finished_plan = self._plan(remote_state, manifest)
        except Exception as e:
            logger.error("~" * 80 + "REMOTE STATE")
            logger.error(remote_state)
            logger.error("~" * 80 + "MANIFEST")
            logger.error(manifest)

            raise e
        self._raise_for_nonconforming_plan(finished_plan)
        return finished_plan

    def apply(self, session, plan: Plan = None):
        if plan is None:
            plan = self.plan(session)

        # TODO: cursor setup, including query tag

        """
            At this point, we have a list of actions as a part of the plan. Each action is one of:
                1. ADD action (CREATE command)
                2. CHANGE action (one or many ALTER or SET PARAMETER commands)
                3. REMOVE action (DROP command, REVOKE command, or a rename operation)
                4. TRANSFER action (GRANT OWNERSHIP command)

            Each action requires:
                • a set of privileges necessary to run commands
                • the appropriate role to execute commands

            Once we've determined those things, we can compare the list of required roles and privileges
            against what we have access to in the session and the role tree.
        """

        session_ctx = data_provider.fetch_session(session)

        _raise_if_plan_would_drop_session_user(session_ctx, plan)

        action_queue = compile_plan_to_sql(session_ctx, plan)
        actions_taken = []

        while action_queue:
            sql = action_queue.pop(0)
            actions_taken.append(sql)
            try:
                if not self._dry_run:
                    execute(session, sql)
            except snowflake.connector.errors.ProgrammingError as err:
                if err.errno == ALREADY_EXISTS_ERR:
                    logger.error(f"Resource already exists: {sql}, skipping...")
                elif err.errno == INVALID_GRANT_ERR:
                    logger.error(f"Invalid grant: {sql}, skipping...")
                elif err.errno == DOES_NOT_EXIST_ERR and sql.startswith("REVOKE"):
                    logger.error(f"Resource does not exist: {sql}, skipping...")
                elif err.errno == DOES_NOT_EXIST_ERR and sql.startswith("DROP"):
                    logger.error(f"Resource does not exist: {sql}, skipping...")
                else:
                    raise err
        return actions_taken

    # def destroy(self, session, manifest: Manifest = None):
    #     session_ctx = data_provider.fetch_session(session)
    #     manifest = manifest or self.generate_manifest(session_ctx)
    #     for urn, data in manifest.items():

    #         if isinstance(data, dict) and data.get("_pointer"):
    #             continue
    #         if urn.resource_type == ResourceType.GRANT:
    #             for grant in data:
    #                 execute(session, lifecycle.drop_resource(urn, grant))
    #         else:
    #             try:
    #                 execute(session, lifecycle.drop_resource(urn, data))
    #             except snowflake.connector.errors.ProgrammingError:
    #                 continue

    def _add(self, resource: Resource):
        if self._finalized:
            raise Exception("Cannot add resources to a finalized blueprint")
        if not isinstance(resource, Resource):
            raise Exception(f"Expected a Resource, got {type(resource)} -> {resource}")
        if resource._finalized:
            raise Exception("Cannot add a finalized resource to a blueprint")
        self._staged.append(resource)

    def add(self, *resources):
        if isinstance(resources[0], list):
            resources = resources[0]
        for resource in resources:
            self._add(resource)


def owner_for_change(change: ResourceChange):
    if change.action == Action.ADD and "owner" in change.after:
        return change.after["owner"]
    elif change.action == Action.CHANGE:
        # TRANSFER actions occur strictly after CHANGE actions, so we use the before owner
        # as the role for the change
        return change.before["owner"]
    elif change.action == Action.REMOVE and "owner" in change.before:
        return change.before["owner"]
    elif change.action == Action.TRANSFER and "owner" in change.before:
        return change.before["owner"]
    else:
        return None


def execution_strategy_for_change(
    change: ResourceChange,
    usable_roles: list[str],
    default_role: str,
    role_privileges: dict[str, list[dict]],
) -> tuple[str, bool]:

    change_is_a_grant = change.urn.resource_type in (
        ResourceType.GRANT,
        ResourceType.FUTURE_GRANT,
        ResourceType.ROLE_GRANT,
    )

    change_is_a_tag_reference = change.urn.resource_type == ResourceType.TAG_REFERENCE

    if change_is_a_grant:

        if change.urn.resource_type == ResourceType.GRANT:
            execution_role = execution_role_for_priv(change.after["priv"])
            if execution_role and execution_role in usable_roles:
                return execution_role, False

        if "SECURITYADMIN" in usable_roles:
            return "SECURITYADMIN", False
        alternate_role = find_role_to_execute_change(change, usable_roles, default_role, role_privileges)
        if alternate_role:
            return alternate_role, False
        raise MissingPrivilegeException(f"{change} requires a role with MANAGE GRANTS privilege")

    elif change_is_a_tag_reference:
        # There are two ways you can create a tag reference:
        # 1. You have the global APPLY TAGS priv on the account (given to ACCOUNTADMIN by default)
        # 2. You have APPLY privilege on the TAG object AND you have ownership of the tagged object
        if "ACCOUNTADMIN" in usable_roles:
            return "ACCOUNTADMIN", False
        alternate_role = find_role_to_execute_change(change, usable_roles, default_role, role_privileges)
        if alternate_role:
            return alternate_role, False
        raise MissingPrivilegeException(f"{change} requires a role with APPLY TAGS privilege")

    change_owner = owner_for_change(change)

    if change_owner:
        # If we can use the owner of the change, use that
        if change_owner in usable_roles and role_can_execute_change(change_owner, change, role_privileges):
            return change_owner, False

        # See if there is another role we can use to execute the change
        alternate_role = find_role_to_execute_change(change, usable_roles, default_role, role_privileges)
        if alternate_role:
            return alternate_role, True
        # This change cannot be executed
        raise MissingPrivilegeException(
            f"Owner {change_owner} does not have the required privileges to execute {change.action} on {change.urn}"
        )

    raise NotImplementedError(change)


# NOTE: this is in a hot loop and maybe should be cached
def role_can_execute_change(
    role: str,
    change: ResourceChange,
    role_privileges: dict[str, list[GrantedPrivilege]],
):

    # Assume ACCOUNTADMIN can do anything
    if role == "ACCOUNTADMIN":
        return True

    if role not in role_privileges:
        return False
    if len(role_privileges[role]) == 0:
        return False

    for granted_priv in role_privileges[role]:
        if granted_priv_allows_change(granted_priv, change):
            return True

    return False


def find_role_to_execute_change(
    change: ResourceChange,
    usable_roles: list[str],
    default_role: str,
    role_privileges: dict[str, list[GrantedPrivilege]],
):
    # Check default role first
    sorted_roles = sorted(usable_roles, key=lambda role: (role != default_role))
    for role in sorted_roles:
        if role_can_execute_change(role, change, role_privileges):
            return role
    return None


def granted_priv_allows_change(granted_priv: GrantedPrivilege, change: ResourceChange):

    # if change.action not in (Action.ADD,):
    #     raise NotImplementedError

    scope = RESOURCE_SCOPES[change.urn.resource_type]

    if isinstance(scope, AccountScope):
        container_name = str(change.urn.account_locator)
    elif isinstance(scope, DatabaseScope):
        container_name = str(change.urn.database().fqn)
    elif isinstance(scope, SchemaScope):
        container_name = str(change.urn.schema().fqn)
    else:
        raise Exception("Exception in granted_priv_allows_change, this should never be reached")

    resource_name = str(change.urn.fqn)

    if change.action == Action.ADD:

        # if resource is a grant, check for MANAGE GRANTS
        if change.urn.resource_type in (ResourceType.GRANT, ResourceType.FUTURE_GRANT, ResourceType.ROLE_GRANT):
            if granted_priv.privilege == AccountPriv.MANAGE_GRANTS:
                return True

        # If resource is a tag reference, check for APPLY TAGS
        elif change.urn.resource_type == ResourceType.TAG_REFERENCE:
            if granted_priv.privilege == AccountPriv.APPLY_TAG:
                return True

        # If we own the resource container, we can always perform ADD
        if is_ownership_priv(granted_priv.privilege) and granted_priv.on == container_name:
            return True

        # If we don't own the container, we need the CREATE privilege for the resource on the container
        create_priv = None
        if change.urn.resource_type in CREATE_PRIV_FOR_RESOURCE_TYPE:
            create_priv = CREATE_PRIV_FOR_RESOURCE_TYPE[change.urn.resource_type]

        if granted_priv.privilege == create_priv and granted_priv.on == container_name:
            return True

        return False

    elif change.action == Action.CHANGE:

        # If we own the resource, we can always make changes
        if is_ownership_priv(granted_priv.privilege) and granted_priv.on == resource_name:
            return True

        # Some resources have a MODIFY privilege that typically allows changes
        if str(granted_priv.privilege) == "MODIFY" and granted_priv.on == resource_name:
            return True

        return False
    elif change.action == Action.REMOVE:
        if is_ownership_priv(granted_priv.privilege) and granted_priv.on == resource_name:
            return True
        return False
    elif change.action == Action.TRANSFER:
        # We must own the resource in order to transfer ownership
        if is_ownership_priv(granted_priv.privilege) and granted_priv.on == resource_name:
            return True
        return False


def sql_commands_for_change(
    change: ResourceChange,
    usable_roles: list[str],
    default_role: str,
    role_privileges: dict[str, list[GrantedPrivilege]],
):
    """
    In Snowflake's RBAC model, a session has an active role, and zero or more secondary roles.

    The active role of a session is set as follows:
    - When a session is started:
        - If the session is configured with a role, that is the active role
        - Otherwise, if the user of the session has a default_role set, and that role exists, that is the active role
        - Otherwise, the PUBLIC role is activated (PUBLIC cannot be revoked)
    - Any time the USE ROLE command is run, the active role is switched

    A session may run any command thats allowed by the active role or any role downstream from it in the role hierarchy.
    When secondary roles are active (by running the command USE SECONDARY ROLES ALL), then the session may also run any
    command that any secondary role or a role downstream from it is allowed to run.

    However, when a CREATE command is run, only the active role is considered. This is because the role that
    creates a new resource owns that resource by default. There are some exceptions with GRANTS.

    For those reasons, we generally don't have to worry about the current role as long as we have activated secondary roles.
    The exception is when creating new resources
    """

    before_change_cmd = []
    change_cmd = None
    after_change_cmd = []

    execution_role, transfer_owner = execution_strategy_for_change(change, usable_roles, default_role, role_privileges)
    before_change_cmd.append(f"USE ROLE {execution_role}")

    if change.action == Action.ADD:
        props = Resource.props_for_resource_type(change.urn.resource_type, change.after)
        change_cmd = lifecycle.create_resource(change.urn, change.after, props)
        if transfer_owner:
            after_change_cmd.append(
                lifecycle.transfer_resource(
                    change.urn,
                    owner=change.after["owner"],
                    copy_current_grants=True,
                )
            )
    elif change.action == Action.CHANGE:
        change_cmd = lifecycle.update_resource(change.urn, change.delta)
    elif change.action == Action.REMOVE:
        change_cmd = lifecycle.drop_resource(
            change.urn,
            change.before,
            if_exists=True,
        )
    elif change.action == Action.TRANSFER:
        change_cmd = lifecycle.transfer_resource(
            change.urn,
            owner=change.after["owner"],
            copy_current_grants=True,
        )

    return before_change_cmd + [change_cmd] + after_change_cmd


def compile_plan_to_sql(session_ctx, plan: Plan):
    sql_commands = []

    sql_commands.append("USE SECONDARY ROLES ALL")
    usable_roles = session_ctx["available_roles"]
    default_role = session_ctx["role"]
    role_privileges = session_ctx["role_privileges"]
    for change in plan:
        # Generate SQL commands
        commands = sql_commands_for_change(
            change,
            usable_roles,
            default_role,
            role_privileges,
        )
        sql_commands.extend(commands)

        # Update state
        if change.action == Action.ADD:
            if change.urn.resource_type == ResourceType.ROLE_GRANT:
                if change.after["to_role"] in usable_roles:
                    usable_roles.append(change.after["role"])
            elif change.urn.resource_type == ResourceType.GRANT:
                grantee_role = change.after["to"]
                if grantee_role not in role_privileges:
                    role_privileges[grantee_role] = []
                for priv in change.after["_privs"]:
                    role_privileges[grantee_role].append(
                        GrantedPrivilege.from_grant(
                            privilege=priv,
                            granted_on=change.after["on_type"],
                            name=change.after["on"],
                        )
                    )
            elif "owner" in change.after:
                # When creating any other resource type, creating implies ownership
                resource_priv_types = PRIVS_FOR_RESOURCE_TYPE[change.urn.resource_type]
                if resource_priv_types and "OWNERSHIP" in resource_priv_types:
                    ownership_priv = PRIVS_FOR_RESOURCE_TYPE[change.urn.resource_type]("OWNERSHIP")
                    owner_role = str(change.after["owner"])
                    if owner_role not in role_privileges:
                        role_privileges[owner_role] = []
                    role_privileges[owner_role].append(
                        GrantedPrivilege.from_grant(
                            privilege=ownership_priv,
                            granted_on=change.urn.resource_type,
                            name=str(change.urn.fqn),
                        )
                    )
                # Special case for databases: if you create a database you own the public schema
                if change.urn.resource_type == ResourceType.DATABASE:
                    role_privileges[owner_role].append(
                        GrantedPrivilege.from_grant(
                            privilege=PRIVS_FOR_RESOURCE_TYPE[ResourceType.SCHEMA]("OWNERSHIP"),
                            granted_on=ResourceType.SCHEMA,
                            name=str(change.urn.fqn) + ".PUBLIC",
                        )
                    )
    return sql_commands


def topological_sort(resource_set: set, references: set):
    # Kahn's algorithm

    # Compute in-degree (# of inbound edges) for each node
    in_degrees = {}
    outgoing_edges = {}

    for node in resource_set:
        in_degrees[node] = 0
        outgoing_edges[node] = set()

    for node, ref in references:
        in_degrees[ref] += 1
        outgoing_edges[node].add(ref)

    # Put all nodes with 0 in-degree in a queue
    queue = Queue()
    for node, in_degree in in_degrees.items():
        if in_degree == 0:
            queue.put(node)

    # Create an empty node list
    nodes = []

    while not queue.empty():
        node = queue.get()
        nodes.append(node)

        # For each of node's outgoing edges
        empty_neighbors = set()
        for edge in outgoing_edges[node]:
            in_degrees[edge] -= 1
            if in_degrees[edge] == 0:
                queue.put(edge)
                empty_neighbors.add(edge)

        # Remove edges to empty neighbors
        outgoing_edges[node].difference_update(empty_neighbors)
    nodes.reverse()
    if len(nodes) != len(resource_set):
        raise Exception("Graph is not a DAG")
    return {value: index for index, value in enumerate(nodes)}



from titan.blueprint import topological_sort

resource_set = {
    'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
    "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    "urn:::role/private_equity_associate",
    'urn:::user/"c.black@arrakisinvestments.com"',
    'urn:::role_grant/private_equity_associate?user="d.gray@arrakisinvestments.com"',
    'urn:::user/"d.gray@arrakisinvestments.com"',
    "urn:::role/private_equity_analyst",
    "urn:::role_grant/private_equity_associate?role=private_equity_analyst",
    "urn:::role_grant/private_equity_associate?role=private_equity_admin",
    "urn:::role/private_equity_admin",
    "urn:::role_grant/private_equity_admin?role=private_equity_associate",
}


refs = [
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        "urn:::role/private_equity_associate",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        'urn:::user/"c.black@arrakisinvestments.com"',
    ),
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        "urn:::role/private_equity_associate",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="c.black@arrakisinvestments.com"',
        'urn:::user/"c.black@arrakisinvestments.com"',
    ),
    (
        'urn:::role_grant/private_equity_associate?user="d.gray@arrakisinvestments.com"',
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="d.gray@arrakisinvestments.com"',
        "urn:::role/private_equity_associate",
    ),
    (
        'urn:::role_grant/private_equity_associate?user="d.gray@arrakisinvestments.com"',
        'urn:::user/"d.gray@arrakisinvestments.com"',
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_analyst",
        "urn:::role/private_equity_analyst",
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_analyst",
        "urn:::role/private_equity_associate",
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_analyst",
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_admin",
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_admin",
        "urn:::role/private_equity_associate",
    ),
    (
        "urn:::role_grant/private_equity_associate?role=private_equity_admin",
        "urn:::role/private_equity_admin",
    ),
    (
        "urn:::role_grant/private_equity_admin?role=private_equity_associate",
        "urn:::account/TITAN_TEST_AWS_ENTERPRISE",
    ),
    (
        "urn:::role_grant/private_equity_admin?role=private_equity_associate",
        "urn:::role/private_equity_associate",
    ),
    (
        "urn:::role_grant/private_equity_admin?role=private_equity_associate",
        "urn:::role/private_equity_admin",
    ),
    ("urn:::role/private_equity_associate", "urn:::account/TITAN_TEST_AWS_ENTERPRISE"),
    ('urn:::user/"c.black@arrakisinvestments.com"', "urn:::account/TITAN_TEST_AWS_ENTERPRISE"),
]


def test_topological_sort():
    sorted_resources = topological_sort(resource_set, set(refs))
    assert len(sorted_resources) == len(resource_set)



from dataclasses import dataclass

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ResourceType
from ..scope import SchemaScope
from ..resource_name import ResourceName

from ..props import Props, IntProp, StringProp


@dataclass(unsafe_hash=True)
class _Sequence(ResourceSpec):
    name: ResourceName
    owner: Role = "SYSADMIN"
    start: int = None
    increment: int = None
    comment: str = None


class Sequence(NamedResource, Resource):
    """
    Description:
        Manages the creation and configuration of sequences in Snowflake, which are objects that generate numeric values according to a specified sequence.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-sequence

    Fields:
        name (string, required): The name of the sequence.
        owner (string or Role): The owner role of the sequence. Defaults to "SYSADMIN".
        start (int): The starting value of the sequence.
        increment (int): The value by which the sequence is incremented.
        comment (string): A comment for the sequence.

    Python:

        ```python
        sequence = Sequence(
            name="some_sequence",
            owner="SYSADMIN",
            start=100,
            increment=10,
            comment="This is a sample sequence."
        )
        ```

    Yaml:

        ```yaml
        sequences:
          - name: some_sequence
            owner: SYSADMIN
            start: 100
            increment: 10
            comment: This is a sample sequence.
        ```
    """

    resource_type = ResourceType.SEQUENCE
    props = Props(
        _start_token="with",
        start=IntProp("start", consume=["with", "="], eq=False),
        increment=IntProp("increment", consume=["by", "="], eq=False),
        comment=StringProp("comment"),
    )
    scope = SchemaScope()
    spec = _Sequence

    def __init__(
        self,
        name: str,
        owner: str = "SYSADMIN",
        start: int = None,
        increment: int = None,
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _Sequence = _Sequence(
            name=self._name,
            owner=owner,
            start=start,
            increment=increment,
            comment=comment,
        )



from dataclasses import dataclass, field

from .resource import Resource, ResourceSpec, NamedResource
from .role import Role
from ..enums import ParseableEnum, ResourceType
from ..resource_name import ResourceName
from ..scope import AccountScope


from ..props import Props, EnumProp, StringProp, BoolProp, StringListProp


class ApiProvider(ParseableEnum):
    AWS_API_GATEWAY = "AWS_API_GATEWAY"
    AWS_PRIVATE_API_GATEWAY = "AWS_PRIVATE_API_GATEWAY"
    AWS_GOV_API_GATEWAY = "AWS_GOV_API_GATEWAY"
    AWS_GOV_PRIVATE_API_GATEWAY = "AWS_GOV_PRIVATE_API_GATEWAY"


@dataclass(unsafe_hash=True)
class _APIIntegration(ResourceSpec):
    name: ResourceName
    api_provider: ApiProvider
    api_aws_role_arn: str
    enabled: bool
    api_allowed_prefixes: list[str]
    api_blocked_prefixes: list[str] = None
    api_key: str = field(default_factory=None, metadata={"fetchable": False})
    owner: Role = "ACCOUNTADMIN"
    comment: str = None


class APIIntegration(NamedResource, Resource):
    """
    Description:
        Manages API integrations in Snowflake, allowing external services to interact with Snowflake resources securely.
        This class supports creating, replacing, and checking the existence of API integrations with various configurations.

    Snowflake Docs:
        https://docs.snowflake.com/en/sql-reference/sql/create-api-integration

    Fields:
        name (string, required): The unique name of the API integration.
        api_provider (string or ApiProvider, required): The provider of the API service. Defaults to AWS_API_GATEway.
        api_aws_role_arn (string, required): The AWS IAM role ARN associated with the API integration.
        api_key (string): The API key used for authentication.
        api_allowed_prefixes (list): The list of allowed prefixes for the API endpoints.
        api_blocked_prefixes (list): The list of blocked prefixes for the API endpoints.
        enabled (bool, required): Specifies if the API integration is enabled. Defaults to TRUE.
        comment (string): A comment or description for the API integration.

    Python:

        ```python
        api_integration = APIIntegration(
            name="some_api_integration",
            api_provider="AWS_API_GATEWAY",
            api_aws_role_arn="arn:aws:iam::123456789012:role/MyRole",
            enabled=True,
            api_allowed_prefixes=["/prod/", "/dev/"],
            api_blocked_prefixes=["/test/"],
            api_key="ABCD1234",
            comment="Example API integration"
        )
        ```

    Yaml:

        ```yaml
        api_integrations:
          - name: some_api_integration
            api_provider: AWS_API_GATEWAY
            api_aws_role_arn: "arn:aws:iam::123456789012:role/MyRole"
            enabled: true
            api_allowed_prefixes: ["/prod/", "/dev/"]
            api_blocked_prefixes: ["/test/"]
            api_key: "ABCD1234"
            comment: "Example API integration"
        ```
    """

    resource_type = ResourceType.API_INTEGRATION
    props = Props(
        api_provider=EnumProp("api_provider", ApiProvider),
        api_aws_role_arn=StringProp("api_aws_role_arn"),
        api_key=StringProp("api_key"),
        api_allowed_prefixes=StringListProp("api_allowed_prefixes", parens=True),
        api_blocked_prefixes=StringListProp("api_blocked_prefixes", parens=True),
        enabled=BoolProp("enabled"),
        comment=StringProp("comment"),
    )
    scope = AccountScope()
    spec = _APIIntegration

    def __init__(
        self,
        name: str,
        api_provider: ApiProvider,
        api_aws_role_arn: str,
        enabled: bool,
        api_allowed_prefixes: list[str],
        api_blocked_prefixes: list[str] = None,
        api_key: str = None,
        owner: str = "ACCOUNTADMIN",
        comment: str = None,
        **kwargs,
    ):
        super().__init__(name, **kwargs)
        self._data: _APIIntegration = _APIIntegration(
            name=self._name,
            api_provider=api_provider,
            api_aws_role_arn=api_aws_role_arn,
            api_key=api_key,
            api_allowed_prefixes=api_allowed_prefixes,
            api_blocked_prefixes=api_blocked_prefixes,
            enabled=enabled,
            owner=owner,
            comment=comment,
        )



import pytest

from titan import resources as res
from titan.enums import ResourceType
from titan.privs import _all_privs_for_resource_type
from titan.identifiers import URN
from titan.resource_name import ResourceName
from titan.resources.resource import ResourcePointer


def test_grant_global_priv():
    grant = res.Grant(priv="CREATE WAREHOUSE", on="ACCOUNT", to="somerole")
    assert grant.priv == "CREATE WAREHOUSE"
    assert grant.on == "ACCOUNT"
    assert grant.to.name == "somerole"
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=CREATE WAREHOUSE&on=account/ACCOUNT"
    assert grant.create_sql() == "GRANT CREATE WAREHOUSE ON ACCOUNT TO SOMEROLE"


def test_grant_account_obj_priv_with_resource():
    wh = res.Warehouse(name="somewh")
    grant = res.Grant(priv="MODIFY", on=wh, to="somerole")
    assert grant.priv == "MODIFY"
    assert grant.on == "SOMEWH"
    assert grant.on_type == ResourceType.WAREHOUSE
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=MODIFY&on=warehouse/SOMEWH"


def test_grant_account_obj_priv_with_kwarg():
    grant = res.Grant(priv="MODIFY", on_warehouse="somewh", to="somerole")
    assert grant.priv == "MODIFY"
    assert grant.on == "SOMEWH"
    assert grant.on_type == ResourceType.WAREHOUSE
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=MODIFY&on=warehouse/SOMEWH"


def test_grant_schema_priv_with_resource():
    sch = res.Schema(name="someschema")
    grant = res.Grant(priv="CREATE VIEW", on=sch, to="somerole")
    assert grant.priv == "CREATE VIEW"
    assert grant.on == "SOMESCHEMA"
    assert grant.on_type == ResourceType.SCHEMA
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=CREATE VIEW&on=schema/SOMESCHEMA"


def test_grant_schema_priv_with_kwarg():
    grant = res.Grant(priv="CREATE VIEW", on_schema="someschema", to="somerole")
    assert grant.priv == "CREATE VIEW"
    assert grant.on == "SOMESCHEMA"
    assert grant.on_type == ResourceType.SCHEMA
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=CREATE VIEW&on=schema/SOMESCHEMA"


def test_grant_all():
    grant = res.Grant(priv="ALL", on_warehouse="somewh", to="somerole")
    assert grant.priv == "ALL"
    assert grant.on == "SOMEWH"
    assert grant.on_type == ResourceType.WAREHOUSE
    assert grant.to.name == "SOMEROLE"
    assert grant._data._privs == _all_privs_for_resource_type(ResourceType.WAREHOUSE)
    assert str(URN.from_resource(grant)) == "urn:::grant/SOMEROLE?priv=ALL&on=warehouse/SOMEWH"


def test_future_grant_schemas_priv():
    grant = res.FutureGrant(priv="CREATE VIEW", on_future_schemas_in_database="somedb", to="somerole")
    assert grant.priv == "CREATE VIEW"
    assert grant.on_type == ResourceType.SCHEMA
    assert grant.in_type == ResourceType.DATABASE
    assert grant.in_name == "SOMEDB"
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::future_grant/SOMEROLE?priv=CREATE VIEW&on=database/SOMEDB.<SCHEMA>"


def test_future_grant_anonymous_target():
    grant = res.FutureGrant(priv="SELECT", on_future_tables_in_schema="someschema", to="somerole")
    assert grant.priv == "SELECT"
    assert grant.on_type == ResourceType.TABLE
    assert grant.in_type == ResourceType.SCHEMA
    assert grant.in_name == "SOMESCHEMA"
    assert grant.to.name == "SOMEROLE"
    assert str(URN.from_resource(grant)) == "urn:::future_grant/SOMEROLE?priv=SELECT&on=schema/SOMESCHEMA.<TABLE>"


def test_future_grant_anonymous_nested_target():
    grant = res.FutureGrant(priv="SELECT", on_future_tables_in_schema="somedb.someschema", to="somerole")
    assert grant.priv == "SELECT"
    assert grant.on_type == ResourceType.TABLE
    assert grant.in_type == ResourceType.SCHEMA
    assert grant.in_name == "somedb.SOMESCHEMA"
    assert grant.to.name == "SOMEROLE"
    assert (
        str(URN.from_resource(grant)) == "urn:::future_grant/SOMEROLE?priv=SELECT&on=schema/SOMEDB.SOMESCHEMA.<TABLE>"
    )


def test_future_grant_referenced_inferred_target():
    schema = res.Schema(name="somedb.someschema")
    grant = res.FutureGrant(priv="SELECT", on_future_tables_in=schema, to="somerole")
    assert grant.priv == "SELECT"
    assert grant.on_type == ResourceType.TABLE
    assert grant.in_type == ResourceType.SCHEMA
    assert grant.in_name == "somedb.SOMESCHEMA"
    assert grant.to.name == "SOMEROLE"
    assert (
        str(URN.from_resource(grant)) == "urn:::future_grant/SOMEROLE?priv=SELECT&on=schema/SOMEDB.SOMESCHEMA.<TABLE>"
    )


def test_role_grant_to_user_with_kwargs():
    grant = res.RoleGrant(role="somerole", to_user="someuser")
    assert grant.role.name == "somerole"
    assert grant._data.to_user is not None
    assert grant.to.name == "someuser"
    assert str(URN.from_resource(grant)) == "urn:::role_grant/SOMEROLE?user=SOMEUSER"


def test_role_grant_to_user_with_resource():
    grant = res.RoleGrant(role="somerole", to=res.User(name="someuser"))
    assert grant.role.name == "somerole"
    assert grant._data.to_user is not None
    assert grant.to.name == "someuser"
    assert str(URN.from_resource(grant)) == "urn:::role_grant/SOMEROLE?user=SOMEUSER"


def test_role_grant_to_role_with_kwargs():
    grant = res.RoleGrant(role="somerole", to_role="someotherrole")
    assert grant.role.name == "somerole"
    assert grant._data.to_role is not None
    assert grant.to.name == "someotherrole"
    assert str(URN.from_resource(grant)) == "urn:::role_grant/SOMEROLE?role=SOMEOTHERROLE"


def test_role_grant_to_role_with_resource():
    grant = res.RoleGrant(role="somerole", to=res.Role(name="someotherrole"))
    assert grant.role.name == "somerole"
    assert grant._data.to_role is not None
    assert grant.to.name == "someotherrole"
    assert str(URN.from_resource(grant)) == "urn:::role_grant/SOMEROLE?role=SOMEOTHERROLE"


def test_grant_redirect_to_all():
    with pytest.raises(ValueError):
        res.Grant(priv="CREATE VIEW", on_all_schemas_in_database="somedb", to="somerole")


def test_grant_on_all():
    grant = res.GrantOnAll(priv="CREATE VIEW", on_all_schemas_in_database="somedb", to="somerole")
    assert grant._data.priv == "CREATE VIEW"
    assert grant._data.on_type == ResourceType.SCHEMA
    assert grant._data.in_type == ResourceType.DATABASE
    assert grant._data.in_name == "somedb"
    assert grant._data.to.name == "somerole"


def test_grant_refs():
    grant = res.Grant.from_sql("GRANT READ ON IMAGE REPOSITORY some_repo TO ROLE titan_app_admin")
    repo = res.ImageRepository(
        name="titan_app_image_repo",
        owner="titan_app_admin",
        database="titan_app_db",
        schema="titan_app",
    )
    assert ResourceName(str(repo.fqn)) == ResourceName("titan_app_db.titan_app.titan_app_image_repo")
    assert grant.priv == "READ"
    assert ResourcePointer("some_repo", ResourceType.IMAGE_REPOSITORY) in grant.refs


def test_grant_priv_is_serialized_uppercase():
    grant = res.Grant(priv="usage", on_warehouse="somewh", to="somerole")
    assert grant.priv == "USAGE"
